import 'dart:async';

import 'package:flutter/widgets.dart';
import 'package:sunny_platform_widgets/sunny_platform_widgets.dart';
import 'package:logging/logging.dart';
import 'package:sunny_core_widgets/sunny_core_widgets.dart';
import 'package:sunny_dart/sunny_dart.dart';
import 'package:sunny_forms/builders/watcher.dart';

final _log = Logger("future_widgets");

/// Shorthand for [Observer] widget.
Widget observe(BuildContext context, String name, WidgetBuilder builder,
        {bool sliver = false, VoidCallback? observation}) =>
    Watcher(
      builder: (_) => builder(_),
      watch: observation ?? () => builder(context),
      name: name,
    );

/// Shorthand for [FutureBuilder] - waits for multiple futures
Widget futures<T>(BuildContext context,
    {required Iterable<FutureOr<T>?> waitFor,
    String? observerName,
    String? debugName,
    bool sliver = false,
    bool awaitAll = false,
    required Widget builder(BuildContext context, List<T>? resolved),
    Widget errorBuilder(BuildContext context, error)?,
    Widget spinner(BuildContext context)?}) {
  return future<List<T>>(
    context,
    key: debugName,
    waitFor: Future.wait<T>(
        waitFor.mapIndexed(((o, idx) => Future.value(o).then((_) {
              _log.info("$debugName: completed $idx");
              return _;
            })) as Future<T> Function(FutureOr<T>?, int)),
        eagerError: true),
    sliver: sliver,
    observerName: observerName,
    builder: (context, resolved) {
      _log.finer(
          "futures: of ${uncapitalize(resolved?.runtimeType.toString() ?? "null")} changed");
      return builder(context, resolved);
    },
    errorBuilder: errorBuilder,
    spinner: spinner,
  );
}

/// Shorthand for [FutureBuilder] - waits for multiple futures
Widget flatFuture<A, B, C>(BuildContext context,
    {FutureOr<A?> a,
    FutureOr<B?> b,
    FutureOr<C?> c,
    String? observerName,
    String? debugName,
    bool sliver = false,
    required Widget builder(BuildContext context, [A? a, B? b, C? c]),
    Widget errorBuilder(BuildContext context, error)?,
    Widget spinner(BuildContext context)?}) {
  return futures<dynamic>(
    context,
    debugName: debugName,
    waitFor: [a, b, c],
    sliver: sliver,
    observerName: observerName,
    builder: (context, resolved) {
      _log.finer(
          "futures: of ${uncapitalize(resolved?.runtimeType.toString() ?? "null")} changed");
      return builder(context, resolved!.tryGet(0) as A?,
          resolved.tryGet(1) as B?, resolved.tryGet(2) as C?);
    },
    errorBuilder: errorBuilder,
    spinner: spinner,
  );
}

extension ListTryGetXX<T> on List<T> {
  T? tryGet(int index) {
    if (index >= length) {
      return null;
    } else {
      return this[index];
    }
  }
}

///// Given a record, will ensure that the data is loaded, and will optionally track changes if [watch] is set
//Widget resolve<V extends Entity>(
//  BuildContext context,
//  Record<V> record,
//  Widget builder(BuildContext context, V resolved), {
//  bool watch = false,
//}) {
//  if (record == null) return builder(context, null);
//  if (!record.isResolved) {
//    return future<V>(context, waitFor: record.future, builder: (context, V data) {
//      return watch
//          ? observe(context, "record-${record.debugName}", (_) {
//              return builder(context, record.value);
//            })
//          : builder(context, record.value);
//    });
//  } else {
//    return watch
//        ? observe(context, "record-${record.debugName}", (_) {
//            return builder(context, record.value);
//          })
//        : builder(context, record.value);
//  }
//}
//
//Widget resolveColumn<V extends Entity>(
//  BuildContext context,
//  Record<V> record,
//  Iterable<Widget> builder(BuildContext context, V resolved),
//) {
//  return future<V>(
//    context,
//    observerName: "record-${uncapitalize(V.toString())}-${record.id}",
//    waitFor: record.future,
//    builder: (context, V data) {
//      return column([...builder(context, record.value)]);
//    },
//  );
//}

Widget futureOr<V>(
  BuildContext context,
  FutureOr<V> waitFor,
  Widget builder(BuildContext context, V? resolved), {
  Widget spinner(BuildContext context)?,
  String? key,
  bool allowNull = false,
}) {
  return future<V>(
    context,
    waitFor: waitFor,
    key: key,
    allowNull: allowNull,
    builder: builder,
    spinner: spinner,
  );
}

/// Shorthand for [FutureBuilder] - combines an [Observer]
Widget future<V>(BuildContext context,
    {FutureOr<V?> waitFor,
    String? observerName,
    bool sliver = false,
    bool allowNull = false,
    V? initialData,
    Duration? timeout,
    String? key,
    required Widget builder(BuildContext context, V? resolved),
    Widget errorBuilder(BuildContext context, error)?,
    Widget spinner(BuildContext context)?}) {
  timeout ??= 20.seconds;
  errorBuilder ??= ((context, err) =>
      sliverBox(ErrorText(err?.toString() ?? "Error"), sliver));
  spinner ??= (context) =>
      sliverBox(Center(child: PlatformCircularProgressIndicator()), sliver);

  initialData ??= (waitFor is V) ? waitFor : null;
  var future = Future.value(waitFor);
  if (initialData == null) future = future.timeout(timeout);
  WidgetBuilder built = (_) => FutureBuilder<V?>(
      key: key == null ? null : Key("$key-future"),
      future: future,
      initialData: initialData,
      builder: (context, snapshot) {
        if (snapshot.hasData || allowNull == true) {
          return builder(context, snapshot.data);
        } else if (snapshot.hasError) {
          _log.severe("Error in future widget: $key: ${snapshot.error}",
              snapshot.error);
          return errorBuilder!(context, snapshot.error);
        } else {
          return spinner!(context);
        }
      });
  return observerName != null
      ? observe(context, observerName, built)
      : built(context);
}

Key keyOf(dyn) {
  if (dyn is Key) {
    return dyn;
  } else {
    return ValueKey(dyn);
  }
}

Widget builder(key, Widget builder(BuildContext context)) {
  return Builder(
    builder: builder,
    key: keyOf(key),
  );
}
