import 'dart:async';

import 'package:flutter/material.dart';
import 'package:sunny_platform_widgets/sunny_platform_widgets.dart';
import 'package:google_maps_webservice/places.dart';
import 'package:http/http.dart';
import 'package:sunny_core_widgets/sunny_core_widgets.dart';
import 'package:sunny_essentials/state.dart';
import 'package:sunny_fluro/sunny_fluro.dart';
import 'package:sunny_forms/header_button.dart';
import 'package:sunny_forms/platform/platform_app_bar_text_field.dart';
import 'package:sunny_forms/sunny_colors.dart';
import 'package:sunny_forms/widget_extensions.dart';
import 'package:sunny_sdk_core/api.dart';

typedef PredictionCallback = void Function(Prediction prediction);
typedef AddressListTileBuilder = Widget Function(
    Prediction prediction, PredictionCallback? onTap);

Future<Prediction?> openAddressPicker(BuildContext context,
    {Location? initialValue,

    /// The types of locations returned by the API
    Set<String>? locationTypes,
    PredictionCallback? onSelect}) async {
  final intl = SunnyIntl.of(context)!;

  return await PlacesAutocomplete.show(
      hint: intl.addressSearch,
      context: context,
      location: initialValue,
      logo: emptyBox,
      types: locationTypes?.toList(),
      apiKey: ThirdPartyCreds.googlePlaces.credentials,
      listTileBuilder: (p, onTap) {
        return PlatformListTile(
          onTap: (_) => onTap!(p),
          leading: Icon(context.icons.location),
          title: Text(p.structuredFormatting?.mainText ?? ''),
          subtitle: Text(p.structuredFormatting?.secondaryText ?? ''),
        );
      },
      language: "en",
      components: []);
}

class PlacesAutocompleteControl extends StatefulWidget {
  final String apiKey;
  final String hint;
  final Location? location;
  final num? offset;
  final num? radius;
  final String? language;
  final String? sessionToken;
  final AddressListTileBuilder? listTileBuilder;
  final List<String>? types;
  final List<Component>? components;
  final bool? strictbounds;
  final Widget? logo;
  final ValueChanged<PlacesAutocompleteResponse>? onError;

  /// optional - sets 'proxy' value in google_maps_webservice
  ///
  /// In case of using a proxy the baseUrl can be set.
  /// The apiKey is not required in case the proxy sets it.
  /// (Not storing the apiKey in the app is good practice)
  final String? proxyBaseUrl;

  /// optional - set 'client' value in google_maps_webservice
  ///
  /// In case of using a proxy url that requires authentication
  /// or custom configuration
  final Client? httpClient;

  PlacesAutocompleteControl(
      {required this.apiKey,
      this.hint = "Search",
      this.offset,
      this.location,
      this.radius,
      this.language,
      this.sessionToken,
      this.types,
      this.components,
      this.strictbounds,
      this.logo,
      this.listTileBuilder,
      this.onError,
      Key? key,
      this.proxyBaseUrl,
      this.httpClient})
      : super(key: key);

  @override
  State<PlacesAutocompleteControl> createState() {
    return _PlacesAutocompleteScaffoldState();
  }

  static PlacesAutocompleteState? of(BuildContext context) =>
      context.findAncestorStateOfType<PlacesAutocompleteState>();
}

class _PlacesAutocompleteScaffoldState extends PlacesAutocompleteState {
  @override
  Widget build(BuildContext context) {
    final appBar = PlatformAppBar(
      leading: HeaderButton.cancel(
        text: "Close",
        isPrimary: true,
        padding: EdgeInsets.only(bottom: 8),
      ),
      title: AppBarPlacesAutoCompleteTextField(
        textController: queryTextController,
        hint: widget.hint,
        focusNode: focusNode,
      ),
    );
    final body = PlacesAutocompleteResult(
      onTap: (prediction) => Navigator.pop(context, prediction),
      logo: widget.logo,
    );
    return PlatformScaffold(
      appBar: appBar,
      body: body,
      key: Key("addressScaffold"),
      widgetKey: Key("addressScaffoldWidget"),
    );
  }
}

class _Loader extends StatelessWidget {
  @override
  Widget build(BuildContext context) {
    return Container(
        constraints: BoxConstraints(maxHeight: 2.0),
        child: LinearProgressIndicator());
  }
}

class PlacesAutocompleteResult extends StatefulWidget {
  final ValueChanged<Prediction>? onTap;
  final Widget? logo;

  PlacesAutocompleteResult({this.onTap, this.logo});

  @override
  _PlacesAutocompleteResult createState() => _PlacesAutocompleteResult();
}

class _PlacesAutocompleteResult extends State<PlacesAutocompleteResult> {
  @override
  Widget build(BuildContext context) {
    final state = PlacesAutocompleteControl.of(context)!;

    if (state.queryTextController!.text.isEmpty ||
        state._response == null ||
        state._response!.predictions.isEmpty) {
      final children = <Widget>[];
      if (state._searching) {
        children.add(_Loader());
      }
      children.add(widget.logo ?? PoweredByGoogleImage());
      return Stack(children: children);
    }
    return PredictionsListView(
      predictions: state._response!.predictions,
      onTap: widget.onTap,
      listTileBuilder: state.widget.listTileBuilder,
    );
  }
}

class AppBarPlacesAutoCompleteTextField extends StatelessWidget {
  final FocusNode? focusNode;
  final TextEditingController? textController;
  final String hint;

  const AppBarPlacesAutoCompleteTextField(
      {Key? key,
      required this.focusNode,
      required this.textController,
      required this.hint})
      : super(key: key);

  @override
  Widget build(BuildContext context) {
    FocusScope.of(context).autofocus(focusNode!);

    return PlatformAppBarTextField(
      focusNode: focusNode,
      controller: textController,
      label: hint,
      iosPadding: SunnyMargins.horiz,
    );
  }
}

class PoweredByGoogleImage extends StatelessWidget {
  final _poweredByGoogleWhite =
      "packages/flutter_google_places/assets/google_white.png";
  final _poweredByGoogleBlack =
      "packages/flutter_google_places/assets/google_black.png";

  @override
  Widget build(BuildContext context) {
    return Row(mainAxisAlignment: MainAxisAlignment.center, children: <Widget>[
      Padding(
          padding: EdgeInsets.all(16.0),
          child: Image.asset(
            Theme.of(context).brightness == Brightness.light
                ? _poweredByGoogleWhite
                : _poweredByGoogleBlack,
            scale: 2.5,
          ))
    ]);
  }
}

class PredictionsListView extends StatelessWidget {
  final List<Prediction> predictions;
  final ValueChanged<Prediction>? onTap;
  final AddressListTileBuilder? listTileBuilder;

  PredictionsListView(
      {required this.predictions, this.onTap, required this.listTileBuilder});

  @override
  Widget build(BuildContext context) {
    return ListView(
      children: predictions.map((Prediction p) {
        return PlaceResultTile(
            prediction: p, onTap: onTap, tileBuilder: listTileBuilder);
      }).toList(),
    );
  }
}

class PlaceResultTile extends StatelessWidget {
  final Prediction prediction;
  final ValueChanged<Prediction>? onTap;
  final AddressListTileBuilder? tileBuilder;

  PlaceResultTile({required this.prediction, this.onTap, this.tileBuilder});

  @override
  Widget build(BuildContext context) {
    return tileBuilder?.call(prediction, onTap) ??
        PlatformListTile(
          leading: Icon(Icons.location_on),
          title: Text(prediction.description!),
          onTap: (context) {
            if (onTap != null) {
              onTap!(prediction);
            }
          },
        );
  }
}

abstract class PlacesAutocompleteState
    extends BaseState<PlacesAutocompleteControl> {
  FocusNode? focusNode;
  TextEditingController? queryTextController;
  PlacesAutocompleteResponse? _response;
  late GoogleMapsPlaces _places;
  late bool _searching;

  @override
  void initState() {
    super.initState();
    focusNode = FocusNode();
    queryTextController = TextEditingController(text: "");

    _places = GoogleMapsPlaces(
        apiKey: widget.apiKey,
        baseUrl: widget.proxyBaseUrl,
        httpClient: widget.httpClient);
    _searching = false;

    queryTextController!.addListener(_onQueryChange);
  }

  @override
  void dispose() {
    _places.dispose();
    queryTextController!.dispose();
//    focusNode.dispose();
    super.dispose();
  }

  Future doSearch(String value) async {
    if (mounted && value.isNotEmpty) {
      setState(() {
        _searching = true;
      });

      final res = await _places.autocomplete(
        value,
        offset: widget.offset,
        location: widget.location,
        radius: widget.radius,
        language: widget.language,
        sessionToken: widget.sessionToken,
        types: widget.types!,
        components: widget.components!,
        strictbounds: widget.strictbounds!,
      );

      if (res.errorMessage?.isNotEmpty == true ||
          res.status == "REQUEST_DENIED") {
        onResponseError(res);
      } else {
        onResponse(res);
      }
    } else {
      onResponse(null);
    }
  }

  void _onQueryChange() {
    doSearch(queryTextController!.text);
  }

  @mustCallSuper
  void onResponseError(PlacesAutocompleteResponse res) {
    if (!mounted) return;

    if (widget.onError != null) {
      widget.onError!(res);
    }
    setState(() {
      _response = null;
      _searching = false;
    });
  }

  @mustCallSuper
  void onResponse(PlacesAutocompleteResponse? res) {
    if (!mounted) return;

    setState(() {
      _response = res;
      _searching = false;
    });
  }
}

class PlacesAutocomplete {
  static Future<Prediction?> show(
      {required BuildContext context,
      required String apiKey,
      String hint = "Search",
      num? offset,
      Location? location,
      num? radius,
      String? language,
      String? sessionToken,
      AddressListTileBuilder? listTileBuilder,
      List<String>? types,
      List<Component>? components,
      bool? strictbounds,
      Widget? logo,
      ValueChanged<PlacesAutocompleteResponse>? onError,
      String? proxyBaseUrl,
      Client? httpClient}) async {
    final builder = (BuildContext ctx) => PlacesAutocompleteControl(
        key: Key("placesAutocomplete"),
        apiKey: apiKey,
        language: language,
        sessionToken: sessionToken,
        components: components,
        types: types,
        location: location,
        radius: radius,
        listTileBuilder: listTileBuilder,
        strictbounds: strictbounds,
        offset: offset,
        hint: hint,
        logo: logo,
        onError: onError,
        proxyBaseUrl: proxyBaseUrl,
        httpClient: httpClient);

    return await Navigator.push<Prediction>(
      context,
      PlatformPageRoute<Prediction>(
        builder: builder,
        settings: PathRouteSettings(
          route: "/addressPicker",
          label: "Address Picker",
        ),
        fullscreenDialog: true,
      ),
    );
  }
}
