import 'dart:convert';

import 'package:flutter/widgets.dart';
import 'package:google_maps_webservice/places.dart' as gmaps;
import 'package:google_maps_webservice/places.dart';
import 'package:http/http.dart' as http;
import 'package:sunny_dart/helpers.dart';
import 'package:sunny_forms/sunny_hud.dart';
import 'package:sunny_sdk_core/sunny_sdk_core.dart';
import 'package:sunny_service_stubs/models.dart';
import 'package:url_launcher/url_launcher.dart';

String photoUrl(reference) {
  final url =
      "https://maps.googleapis.com/maps/api/place/photo?maxwidth=200&maxheight=200&photoreference=$reference&key=${ThirdPartyCreds.googlePlaces.credentials}";
  return url;
}

Future openMap(BuildContext context, gmaps.PlaceDetails placeDetails) async {
  final geo = placeDetails.geometry!.location;

  if (infoX.isIOS) {
    final url = "https://maps.apple.com/?q=${geo.lat},${geo.lng}";
    if (await canLaunch(url)) {
      await launch(url);
    } else {
      SunnyHud.error(context, "Unable to show a map at this time");
    }
  } else {
    SunnyHud.info(context, "Coming soon");
  }
}

extension SunnyPlaceFromGoogleMapsExt on SunnyPlace {
  void fromGooglePlaceDetails(PlaceDetails details, {String? placeId}) {
    final location = this;
    final keyedDetails = details.addressComponents
        .expand(
            (comp) => comp.types.map((type) => MapEntry(type, comp.shortName)))
        .toList()
        .asMap()
        .map((_, entry) => entry);

    location.googlePlaceId = placeId;
    location.streetLineOne = Lists.compactEmpty([
      keyedDetails["street_number"],
      keyedDetails["route"],
    ]).join(" ");

    if (!location.streetLineOne!.contains(details.name)) {
      location.description = details.name;
    }

    location.region = keyedDetails["administrative_area_level_1"];
    location.locality = keyedDetails["locality"];
    location.countryCode = keyedDetails["country"];
    location.postalCode = keyedDetails["postal_code"];
  }

  void fromGooglePrediction(Prediction prediction) {
    final location = this;
    location.googlePlaceId = prediction.placeId;

    location.description = prediction.structuredFormatting!.mainText;
    location.type = prediction.structuredFormatting!.secondaryText;
  }
}

Future openMapForLocation(
    BuildContext context, IPhysicalLocation sunnyPlace) async {
  String? googlePlaceId = sunnyPlace.googlePlaceId;
  if (googlePlaceId == null) {
    GooglePlaceId? lookupResult;
    await SunnyHud.spinner(context, "Calculating location",
        operation: () async {
      lookupResult = await lookupAddressInGoogle(sunnyPlace);
      return lookupResult;
    }).value;
    googlePlaceId = lookupResult?.placeId;
  }

  final details = await lookupPlaceInGoogle(googlePlaceId);
  illegalState("Need to fix this");
  // sunnyPlace.placeDetails = details;

  if (details == null) {
    await SunnyHud.error(context, "Unable to generate map to this location")
        .value;
    return;
  }

  final geo = details.geometry!.location;

  if (infoX.isIOS) {
    final url = "https://maps.apple.com/?q=${geo.lat},${geo.lng}";
    if (await canLaunch(url)) {
      await launch(url);
    } else {
      SunnyHud.error(context, "Unable to show a map at this time");
    }
  } else {
    SunnyHud.info(context, "Coming soon");
  }
}

Future<gmaps.PlaceDetails?> lookupPlaceInGoogle(String? placeId) async {
  if (placeId == null) {
    return null;
  }
  final placesApi =
      gmaps.GoogleMapsPlaces(apiKey: ThirdPartyCreds.googlePlaces.credentials);
  var gmapsUrl = placesApi.buildDetailsUrl(placeId: placeId);
  var url = rproxy.gmaps(gmapsUrl);
  var resp = await http.get(Uri.parse(url));
  return gmaps.PlacesDetailsResponse.fromJson(
          json.decode(resp.body) as Map<String, dynamic>)
      .result;
}

Future<List<IPhysicalLocation>> searchPlaceInGoogle(String queryString) async {
  if (queryString.isEmpty) {
    return [];
  }
  final placesApi =
      gmaps.GoogleMapsPlaces(apiKey: ThirdPartyCreds.googlePlaces.credentials);
  final dev = await infoX.deviceInfo;

  final location = fromGeoPoint(dev.geo);

  final url = location == null
      ? placesApi.buildAutocompleteUrl(input: queryString)
      : placesApi.buildAutocompleteUrl(
          input: queryString,
          location: location,
          origin: location,
        );
  final purl = rproxy.gmaps(url);
  var resp = await http.post(Uri.parse(purl), body: <String, dynamic>{});
  final response = PlacesAutocompleteResponse.fromJson(
      json.decode(resp.body) as Map<String, dynamic>);
  assert(!response.isInvalid, "Response was invalid");
  assert(!response.isDenied, "Response was invalid");
  assert(!response.isOverQueryLimit, "Response was invalid");
  return response.predictions
      .map((e) => IPhysicalLocation.of()..fromGooglePrediction(e))
      .toList();
}

class GooglePlaceId {
  final String placeId;

  GooglePlaceId(this.placeId);
}

Location? fromGeoPoint(GeoPoint? point) {
  if (point?.lat == null || point?.lon == null) {
    return null;
  } else {
    return Location(lat: point!.lat!, lng: point.lon!);
  }
}

Future<GooglePlaceId?> lookupAddressInGoogle(SunnyPlace place) async {
  final address = addressToString(place)!.join(" ");
  final placesApi =
      gmaps.GoogleMapsPlaces(apiKey: ThirdPartyCreds.googlePlaces.credentials);
  final response = await placesApi.searchByText(address);
  if (response.results.isNotEmpty != true) {
    return null;
  } else {
    for (final result in response.results) {
      return GooglePlaceId(result.placeId);
    }
  }
  return null;
}

extension RProxyServerGMaps on RProxyServer {
  RProxy get gmaps =>
      this.proxy("gmaps", "https://maps.googleapis.com/maps/api");
}
