import 'dart:async';
import 'dart:io';
import 'dart:typed_data';

import 'package:cached_network_image/cached_network_image.dart';
import 'package:cached_network_image_platform_interface/cached_network_image_platform_interface.dart';
import 'package:flutter/material.dart';
import 'package:flutter/widgets.dart';
import 'package:logging/logging.dart';
import 'package:pfile_picker/pfile_picker.dart';
import 'package:sunny_core_widgets/sunny_core_widgets.dart';
import 'package:sunny_dart/extensions.dart';
import 'package:sunny_dart/ui.dart';
import 'package:sunny_forms/icon/icon_info.dart';
import 'package:sunny_forms/icon/stream_image_provider.dart';
import 'package:sunny_forms/media/choosers.dart';
import 'package:sunny_forms/widget_extensions.dart';
import 'package:sunny_service_stubs/models.dart';
import 'package:sunny_service_stubs/services.dart';

/// Custom logger due to isolate use.
final _log = Logger("image_widgets");

/// The default image size
const double kDefaultImageSize = 40;

/// Loads based on a URL, file, etc - if the url starts with assets:// then the image is loaded from assets.  Also, if the
/// image is from the sunny api server, it will attach the auth header if it's present.
///
/// It's usually easier to use [
ImageProvider? imageProviderOf(final dynamic source,
    {bool useHttpGetForWeb = true, void errorCallback()?}) {
  if (source == null) return null;
  if (source is File) return FileImage(source);
  if (source is PlatformFile)
    return StreamImageProvider.ofPFile(PFile.of(source)!);
  if (source is PFile) return StreamImageProvider.ofPFile(source);
  if (source is Uint8List) return MemoryImage(source);
  if (source is ChosenMedia) {
    if (source.content != null) {
      return imageProviderOf(
          (source.content as IImageContent?)?.imageUrl?.toUri());
    } else if (source.file != null) {
      return StreamImageProvider.ofPFile(source.file!);
    }
  }

  String imageUrl;
  if (source is Uri) {
    imageUrl = "$source";
  } else if (source is IconInfo) {
    if (source.type == IconType.named_) return null;
    imageUrl = source.icon;
  } else if (source is String) {
    imageUrl = source;
  } else {
    _log.info(
        "IMAGE: WARN ->  UNKNOWN SOURCE $source is ${source.runtimeType}");
    imageUrl = "$source";
  }
  if (imageUrl.startsWith("assets://")) {
    return AssetImage(
      "assets/${imageUrl.replaceFirst("assets://", "")}",
    );
  } else {
//    final apiClient = BearerApiClient.instance;
//    final headers = {
//      if (imageUrl.startsWith(apiClient.config.sunnyBaseUrl))
//        "Authorization": apiClient.authHeader,
//    };
    return CachedNetworkImageProvider(imageUrl,
        imageRenderMethodForWeb: useHttpGetForWeb
            ? ImageRenderMethodForWeb.HttpGet
            : ImageRenderMethodForWeb.HttpGet,
//        headers: headers,
        // ignore: deprecated_member_use
        errorListener: errorCallback);
  }
}

/// Loads an image widget from a variety of known sources: Uri, String, File, bytes.  The placeholder
/// is only used when the imageSource is a network source (Uri, String).  In that case, the placeholder
/// is displayed while the image is loading.
Widget? imageWidgetOf(
  BuildContext context,
  final imageSource, {
  Key? key,
  bool circle = true,
  double size = kDefaultImageSize,
  Widget? placeholder,
  Color? background,
  bool fallbackToPlaceholder = true,
  iconOverlay,
}) {
  if (imageSource is Iterable) {
    final Iterable allOptions = [
      ...imageSource,
      if (fallbackToPlaceholder && placeholder != null) placeholder,
    ].where((_) => _ != null);
    final eachImageOption = allOptions.map((source) {
      return imageWidgetOf(
        context,
        source,
        circle: circle,
        key: key,
        size: size,
        placeholder: placeholder,
        fallbackToPlaceholder: false,
        iconOverlay: iconOverlay,
      );
    });
    return eachImageOption.firstOrNull;
  }

  Widget? widget;

  if (imageSource is IconInfo) {
    if (imageSource.type == IconType.named_) {
      widget = context
          .findIcon(imageSource.icon)
          ?.copyWith(
            size: size,
            color: imageSource.background.toColor(),
            circular: imageSource.circular,
          )
          .widget;
      if (widget == null) {
        _log.severe(
            "Missing named icon ${imageSource.icon}", null, StackTrace.current);
      }
    }
    background ??= imageSource.backgroundColor;
  } else if (imageSource is NamedIcon) {
    widget = imageSource
        .copyWith(size: size, circular: circle, color: background)
        .icon;
  } else if (imageSource is IconData) {
    widget = Icon(imageSource, size: size, color: background);
  } else if (imageSource is Icon) {
    widget = Icon(imageSource.icon, size: size, color: background);
  }
  if (widget is Icon && circle != true) {
    /// Let it pass through
  } else if (imageSource is Widget) {
    widget = _widgetImage(imageSource,
        key: key, circle: circle, size: size, background: background);
  } else if (widget == null) {
    final ImageProvider? data = imageProviderOf(imageSource);
    final fallback = fallbackToPlaceholder ? placeholder : null;
    if (data == null && fallback == null) return null;
    if (data is CachedNetworkImageProvider) {
      widget = _networkImage(data, placeholder,
          key: key, background: background, circle: circle, size: size);
    } else {
      widget = _memoryImage(data,
          key: key,
          fallback: fallback,
          background: background,
          circle: circle,
          size: size);
    }
  }

  if (iconOverlay != null) {
    // Figure out appropriate offset padding - we have to make the containing box slightly larger to accommodate for
    // the overlay icon needing to extend further (to not overlay as much)
    final padding = EdgeInsets.only(
        bottom: (size ~/ 15).roundToDouble(),
        right: (size ~/ 15).roundToDouble());
    return Stack(
      alignment: AlignmentDirectional.bottomEnd,
      children: [
        Container(
          padding: padding,
          child: widget,
        ),
        if (iconOverlay is NamedIcon) iconOverlay.build(size: size / 3),
        if (iconOverlay is! NamedIcon) iconOverlay as Widget,
      ],
    );
  } else {
    return widget;
  }
}

Widget _networkImage(CachedNetworkImageProvider image, Widget? placeholder,
    {Key? key,
    bool circle = true,
    Color? background,
    BoxFit fit = BoxFit.cover,
    double? padding,
    double size = kDefaultImageSize}) {
  padding ??= 0;
  final networkImage = PlatformNetworkImage(
    image.url,
    key: key,
    httpHeaders: image.headers,
    fit: fit,
    placeholder: (context, _) => placeholder!,
    errorWidget: (context, url, err) => placeholder!,
    width: size,
    height: size,
  );
  return circle ? ClipOval(child: networkImage) : networkImage;
}

Widget _memoryImage(
  ImageProvider? image, {
  Key? key,
  bool circle = true,
  Widget? fallback,
  Color? background,
  double size = kDefaultImageSize,
}) {
  return Container(
    key: key,
    height: size,
    width: size,
    decoration: BoxDecoration(
      color: background,
      shape: circle ? BoxShape.circle : BoxShape.rectangle,
      image: image == null
          ? null
          : DecorationImage(
              fit: BoxFit.fill,
              image: image,
            ),
    ),
    child: ClipRect(child: fallback),
  );
}

/// Uses an arbitrary widget as the image.
Widget _widgetImage(
  Widget image, {
  Key? key,
  bool circle = true,
  Color? background,
  double size = kDefaultImageSize,
}) {
  final container = Container(
    key: key,
    height: size,
    width: size,
    color: background,
    child: circle ? image : ClipRect(child: image),
  );
  if (circle) {
    return ClipOval(child: container);
  } else {
    return container;
  }
}

FutureOr precacheImageIfNetwork(imageUrl, BuildContext context) {
  final imageData = imageProviderOf(imageUrl);
  if (imageData is CachedNetworkImageProvider) {
    return precacheImage(imageData, context);
  }
  if (imageUrl is! String) imageUrl = "$imageUrl";
  if (!imageUrl.toString().startsWith("assets://")) {
    return precacheImage(
        CachedNetworkImageProvider(imageUrl.toString()), context);
  }
}
