import 'dart:math';
import 'dart:typed_data';

import 'package:crop/crop.dart';
import 'package:extended_image/extended_image.dart';
import 'package:flutter/material.dart';
import 'package:image/image.dart' as img;
import 'package:pfile/web/raw_pfile.dart';
import 'package:pfile_picker/pfile_picker.dart';
import 'package:sunny_dart/sunny_dart.dart';
import 'package:sunny_forms/dividing_line.dart';
import 'package:sunny_forms/platform/platform_flat_button.dart';
import 'package:sunny_forms/platform/platform_raised_button.dart';
import 'package:sunny_forms/sunny_hud.dart';

class CropImagePage extends StatefulWidget {
  final Uint8List source;
  final String sourceName;
  final double cropAspectRatio;

  const CropImagePage(
      {Key? key,
      required this.sourceName,
      this.cropAspectRatio = 1.0,
      required this.source})
      : super(key: key);

  CropImagePage.ofPlatformFile({
    Key? key,
    required PFile source,
    Uint8List? bytes,
    this.cropAspectRatio = 1.0,
    // ignore: deprecated_member_use
  })  : source = bytes ?? source.bytes ?? illegalState("Need loaded bytes"),
        sourceName = source.name!,
        super(key: key);

  @override
  _CropImagePageState createState() => _CropImagePageState();
}

class _CropImagePageState extends State<CropImagePage> {
  bool _isSubmitting = false;
  late Size _boxSize;
  final editorKey =
      GlobalKey<ExtendedImageEditorState>(debugLabel: 'imageEditor');

  @override
  Widget build(BuildContext context) {
    return Scaffold(
      body: Column(
        children: <Widget>[
          Expanded(
            child: Container(
              color: Colors.black,
              padding: EdgeInsets.all(8),
              child: LayoutBuilder(
                builder: (_, layout) {
                  _boxSize = layout.biggest;
                  return ExtendedImage.memory(
                    widget.source,
                    extendedImageEditorKey: editorKey,
                    fit: BoxFit.contain,
                    mode: ExtendedImageMode.editor,
                    initEditorConfigHandler: (state) {
                      return EditorConfig(
                        cropAspectRatio: widget.cropAspectRatio,
                      );
                    },
                  );
                },
              ),
            ),
          ),
          SizedBox(
            height: 50,
            child: Layout.row().alignCenter.build([
              PlatformRaisedButton(
                child: textOrNull(_isSubmitting ? "Processing..." : "Save",
                    color: Colors.white),
                color: sunnyColors.primaryColor,
                onPressed: () async {
                  setState(() {
                    _isSubmitting = true;
                  });

                  try {
                    Future.microtask(
                      () async {
                        var editState = editorKey.currentState!;
                        final ext = widget.sourceName.extension ?? 'jpg';
                        final name = widget.sourceName.trimEnd(".$ext");
                        var newName = name;

                        var scale = editState.editAction?.totalScale ?? 1.0;

                        if (scale != 1.0) {
                          newName += "_scale_${NumXX(scale).roundTo(2)}";
                        }

                        var uiOffset =
                            editState.editAction?.layoutTopLeft ?? Offset.zero;

                        if (uiOffset != Offset.zero) {
                          newName +=
                              "_offset_${uiOffset.dx.round()}x${uiOffset.dy}";
                        }
                        newName += '.$ext';
                        Navigator.pop(
                            context,
                            RawPFile.ofBytes(newName,
                                Uint8List.fromList(editState.rawImageData)));
                      },
                    );
                  } catch (e) {
                    SunnyHud.error(context, "Error: $e");
                    Navigator.pop(context);
                  }
                },
              ),
            ]),
          ),
        ],
      ),
    );
  }
}

class CropWidgetControls extends StatefulWidget {
  final CropController? controller;

  const CropWidgetControls({Key? key, required this.controller})
      : super(key: key);

  @override
  _CropWidgetControlsState createState() => _CropWidgetControlsState();
}

class _CropWidgetControlsState extends State<CropWidgetControls> {
  double _slider = 1.0;

  @override
  Widget build(BuildContext context) {
    return SizedBox(
      height: 50,
      child: Row(
        children: <Widget>[
          IconButton(
            icon: Icon(Icons.undo),
            tooltip: 'Undo',
            onPressed: () {
              widget.controller!.rotation = 0;
              widget.controller!.scale = 1;
              widget.controller!.offset = Offset.zero;
              setState(() {
                _slider = 1.0;
              });
            },
          ),
          Expanded(
            child: Slider.adaptive(
              divisions: 100,
              value: _slider,
              min: 1.0,
              max: 3.0,
              label: '$_slider°',
              onChanged: (n) {
                _slider = n;
                setState(() {
                  _slider = n;
                });
                print("Setting scale: $_slider");
                widget.controller!.scale = _slider;
              },
            ),
          ),
        ],
      ),
    );
  }
}
