import 'dart:async';

import 'package:flutter/foundation.dart';
import 'package:flutter/material.dart';
import 'package:flutter/widgets.dart';
import 'package:logging/logging.dart';
import 'package:multi_image_picker/multi_image_picker.dart';
import 'package:pfile_picker/pfile_picker.dart';
import 'package:sunny_dart/sunny_dart.dart';
import 'package:sunny_forms/media/file_source_type.dart';
import 'package:sunny_forms/media/media_content_type.dart';
import 'package:sunny_forms/media/media_url_form.dart';
import 'package:sunny_forms/options.dart';
import 'package:sunny_service_stubs/models.dart';

final _log = Logger("choosers");

class ChosenMedia {
  /// When the result is a locally selected file
  final PFile? file;

  /// When the selection is already uploaded
  final IContent? content;

  const ChosenMedia.ofFile(this.file) : content = null;
  const ChosenMedia.ofContent(this.content) : file = null;
}

class Choosers {
  static Future<List<ChosenMedia>> flutterFilePicker(
    BuildContext context, {
    int? maxSelections = 1,
    VoidCallback? onLoading,
    FlutterMediaContentType? mediaType,

    /// By default will show all available
    Set<FileSourceType>? sourceTypes,
    dynamic extraOptions,
  }) async {
    List<ChosenMedia> res;
    var hasStartedReading = false;

    var fileType = mediaType == FlutterMediaTypes.videoExt
        ? FileType.video
        : mediaType == FlutterMediaTypes.imageExt
            ? FileType.image
            : FileType.any;

    sourceTypes ??= {
      const FileSourceType.files(),
      if (!kIsWeb) const FileSourceType.gallery(),
      const FileSourceType.url(),
    };
    assert(sourceTypes.length > 0);

    FileSourceType? sourceType = sourceTypes.length == 1
        ? sourceTypes.first
        : (await showPicker(context,
                options: {
                  for (var option in sourceTypes) option.toOption(),
                },
                extraOptions: extraOptions))
            ?.value;

    if (sourceType == null) {
      return [];
    }

    if (sourceType == const FileSourceType.files()) {
      final sel = await FilePicker.platform.pickFiles(
          type: FileType.custom,
          allowedExtensions: mediaType!.acceptedFileTypes,
          dialogTitle: "Choose an image",
          withData: false,
          // allowCompression: true,
          withReadStream: true,
          allowCompression: true,
          allowMultiple: maxSelections! > 1,
          onFileLoading: (_) {
            if (!hasStartedReading) {
              hasStartedReading = true;
              onLoading?.call();
            }
          });
      res = [...?sel?.files.map((file) => ChosenMedia.ofFile(file))];
    } else if (sourceType == const FileSourceType.gallery()) {
      final sel = await FilePicker.platform.pickFiles(
          type: fileType,
          // allowCompression: true,
          withReadStream: true,
          allowCompression: true,
          allowMultiple: maxSelections! > 1,
          onFileLoading: (_) {
            if (!hasStartedReading) {
              hasStartedReading = true;
              onLoading?.call();
            }
            _log.info("File is loading... $_");
          });

      res = [...?sel?.files.map((file) => ChosenMedia.ofFile(file))];
    } else if (sourceType == const FileSourceType.url()) {
      final sel = await showMediaUrlForm(context, mediaType: mediaType);
      res = [
        if (sel != null) ChosenMedia.ofContent(sel as IContent),
      ];
    } else {
      throw "Illegal state.  Invalid choice: ${sourceType}";
    }

    return res;
  }

  @deprecated
  static Future<List<ChosenMedia>> chooseImages(
    BuildContext context, {
    int? maxSelections = 1,
    VoidCallback? onLoading,
    FlutterMediaContentType? mediaType,

    /// By default will show all available
    Set<FileSourceType>? sourceTypes,
    dynamic extraOptions,
  }) async {
    final media = await MultiImagePicker.pickImages(
        maxImages: maxSelections!,
        enableCamera: true,
        materialOptions: MaterialOptions(
          textOnNothingSelected: "Nothing was selected!",
          autoCloseOnSelectionLimit: true,
          lightStatusBar: true,
        ),
        cupertinoOptions: CupertinoOptions(
          autoCloseOnSelectionLimit: true,
        ));

    if (media.isNotEmpty != true) {
      return [];
    }

    const double max = 1800.0;
    final images = await Future.wait(media.map((asset) async {
      final w = asset.originalWidth!;
      final h = asset.originalHeight!;
      double ratio, ww, hh;
      if (asset.isLandscape) {
        if (w > max) {
          ratio = max / w;
        } else {
          ratio = 1;
        }
      } else {
        if (h > max) {
          ratio = max / h;
        } else {
          ratio = 1;
        }
      }
      ww = ratio * w;
      hh = ratio * h;
      final image = await asset.getThumbByteData(ww.truncate(), hh.truncate());
      final buffer = image.buffer;
      final bytes =
          buffer.asUint8List(image.offsetInBytes, image.lengthInBytes);
      return PFile.of(bytes, name: asset.name, size: bytes.length);
    }));

    return images.mapNotNull((item) => ChosenMedia.ofFile(item));
  }
}
