import 'package:chewie/chewie.dart';
import 'package:equatable/equatable.dart';
import 'package:flutter/material.dart';
import 'package:muse_ai/muse_ai.dart';
import 'package:sunny_dart/sunny_dart.dart';
import 'package:sunny_forms/dividing_line.dart';
import 'package:sunny_forms/form_fields/media_source.dart';
import 'package:sunny_forms/icon/image_widgets.dart';
import 'package:sunny_forms/muse.dart';
import 'package:sunny_forms/players/stateful_video_player.dart';
import 'package:sunny_sdk_core/model/media_service.dart';
import 'package:sunny_service_stubs/models.dart';
import 'package:sunny_service_stubs/services/media_types.dart';

class _FlutterMediaTypes with LoggingMixin {
  final _types = <String, FlutterMediaContentType>{};
  final _typesBySubject = <Type, FlutterMediaContentType>{};
  final _typesByExtension = <String, Set<FlutterMediaContentType>>{};

  _FlutterMediaTypes._();

  FlutterMediaContentType<T> getForInstance<T extends Object>(T instance) {
    return (_typesBySubject[T] ??
            (throw "No media type for ${instance.runtimeType}"))
        as FlutterMediaContentType<T>;
  }

  FlutterMediaContentType<T> getByName<T extends Object>(String name) {
    return (_types[name] ?? (throw "No media type for ${name}"))
        as FlutterMediaContentType<T>;
  }

  FlutterMediaContentType<T> get<T extends Object>() {
    return (_typesBySubject[T] ?? (throw "No media type for ${T}"))
        as FlutterMediaContentType<T>;
  }

  void register<T extends Object>(FlutterMediaContentType<T> type) {
    if (_types.containsKey(type.name)) {
      log.warning("Type already registered: ${type.name}");
    }
    _types[type.name] = type;
    _typesBySubject[T] = type;

    // Register all extensions
    type.acceptedFileTypes.forEach((extension) {
      _typesByExtension.putIfAbsent(extension, () => {}).add(type);
    });
  }
}

final FlutterMediaTypes = _FlutterMediaTypes._();

abstract class BaseFlutterMediaContentType<F extends Object>
    extends FlutterMediaContentType<F> with EquatableMixin {
  @override
  List get props => [name];

  const BaseFlutterMediaContentType();
}

/// Represents either image or video content, allows for different upload/preview
abstract class FlutterMediaContentType<F extends Object>
    implements MediaContentType<F> {
  const FlutterMediaContentType();

  Color get barrierColor;

  Widget buildPreview(BuildContext context, MediaSourceData source,
      {Widget? child, required bool isUploading, double? borderRadius});

  Widget buildThumbnail(
    BuildContext context,
    MediaSourceData source, {
    Widget? child,
    required bool isUploading,
    double? borderRadius,
  });

  String get title => fileType.capitalize();
}

abstract class FlutterMediaContentTypeMixin<F extends Object>
    implements FlutterMediaContentType<F> {
  String get title => fileType.capitalize();
}

class FlutterVideoContentType extends VideoContentType
    with FlutterMediaContentTypeMixin<IVideoContent> {
  const FlutterVideoContentType._() : super.extended();

  @override
  Widget buildPreview(BuildContext context, MediaSourceData media,
      {Widget? child, required bool isUploading, double? borderRadius}) {
    final player = buildVideo(
      context,
      media,
      child: child,
      isUploading: isUploading,
      borderRadius: borderRadius,
    );
    return Stack(children: [
      Container(
        decoration: BoxDecoration(
          color: sunnyColors.g100,
          borderRadius: BorderRadius.all(Radius.circular(borderRadius ?? 10)),
        ),
      ),
      player,
      if (child != null && isUploading) child,
    ]);
  }

  Widget buildVideo(BuildContext context, MediaSourceData media,
      {Widget? child,
      bool? isUploading,
      double? borderRadius,
      Consumer<ChewieController>? controller}) {
    ChewieController? _chewie;
    Widget player;
    if (media.uri!.startsWith("https://muse.ai")) {
      try {
        final museIndex = sunny.get<MuseIndex>();
        final args = MuseArgs.ofUri(media.uri!);
        final video = museIndex.findVideoBySvid(args.museId)!;
        player = StatefulVideoPlayer(
          key: Key("videoUrl${media.uri}"),
          videoUrl: video.format("mp4"),
          thumbnailUrl: null,
          // isThumbnail: true,
          showControls: true,
          autoplay: false,
          initializer: (chewie) {
            _chewie ??= chewie;
            controller?.call(chewie!);
            if (args.start > 0) {
              chewie!.videoPlayerController
                  .seekTo(Duration(seconds: args.start));
            }
          },
          onCompletion: () {},
        );
      } catch (e) {
        rethrow;
      }
    } else {
      player = media.uri != null
          ? StatefulVideoPlayer(
              key: Key("videoUrl${media.uri}"),
              videoUrl: media.uri,
              thumbnailUrl: null,
              // isThumbnail: true,
              showControls: true,
              autoplay: false,
              initializer: (chewie) {
                _chewie ??= chewie;
              },
              onCompletion: () {},
            )
          : StatefulVideoPlayer(
              key: Key("videoFile${media.file?.path ?? media.file?.name}"),
              videoFile: media.file,
              // isThumbnail: true,
              thumbnailUrl: null,
              autoplay: false,
              showControls: true,
              initializer: (chewie) {
                _chewie ??= chewie;
              });
    }
    return player;
  }

  @override
  Widget buildThumbnail(BuildContext context, MediaSourceData media,
      {Widget? child, bool? isUploading, double? borderRadius}) {
    ChewieController? chewie;
    final player = buildVideo(
      context,
      media,
      child: child,
      isUploading: isUploading,
      borderRadius: borderRadius,
      controller: (_) => chewie = _,
    );
    return Tappable(
      onTap: (_) async {
        chewie?.enterFullScreen();
        chewie?.play();
      },
      child: Stack(
        alignment: Alignment.center,
        children: [
          Container(
            decoration: BoxDecoration(
              color: sunnyColors.g100,
              borderRadius:
                  BorderRadius.all(Radius.circular(borderRadius ?? 10)),
            ),
          ),
          player,
          if (child != null && isUploading!) child,
        ],
      ),
    );
  }

  @override
  Color get barrierColor => Colors.white;
}

class FlutterImageContentType extends ImageContentType
    with FlutterMediaContentTypeMixin<IImageContent> {
  const FlutterImageContentType._() : super.extended();

  @override
  Widget buildThumbnail(BuildContext context, MediaSourceData source,
      {Widget? child, required bool isUploading, double? borderRadius}) {
    return buildPreview(context, source,
        child: child, isUploading: isUploading, borderRadius: borderRadius);
  }

  @override
  Widget buildPreview(BuildContext context, MediaSourceData media,
      {Widget? child, required bool isUploading, double? borderRadius}) {
    if (isUploading) {
      return Container(
        decoration: BoxDecoration(
          image: DecorationImage(
            image: imageProviderOf(media.file)!,
            colorFilter: ColorFilter.mode(
                Colors.white.withOpacity(0.6), BlendMode.lighten),
            fit: BoxFit.cover,
          ),
          borderRadius: BorderRadius.all(Radius.circular(borderRadius ?? 10)),
        ),
        child: child,
      );
    } else {
      return Container(
        decoration: BoxDecoration(
          image: media.uri == null
              ? null
              : DecorationImage(
                  image: imageProviderOf(media.uri)!,
                  fit: BoxFit.cover,
                ),
          borderRadius: BorderRadius.all(Radius.circular(borderRadius ?? 10.0)),
        ),
      );
    }
  }

  @override
  Color get barrierColor => sunnyColors.barrierColor;
}

extension FlutterMediaTypesCoreTypes on _FlutterMediaTypes {
  FlutterMediaContentType<IImageContent> get imageExt =>
      const FlutterImageContentType._();
  FlutterMediaContentType<IVideoContent> get videoExt =>
      const FlutterVideoContentType._();
}
