import 'dart:async';

import 'package:flutter/material.dart';
import 'package:flutter/widgets.dart';
import 'package:logging/logging.dart';
import 'package:multi_image_picker/multi_image_picker.dart';
import 'package:permission_handler/permission_handler.dart';
import 'package:sunny_dart/sunny_dart.dart';
import 'package:sunny_forms/media/media_content_type.dart';
import 'package:sunny_forms/permissions.dart';
import 'package:sunny_forms/sunny_hud.dart';

import 'choosers.dart';
import 'crop.dart';
import 'file_source_type.dart';
import 'platform_files.dart';

final _log = Logger("mediaPicker");

typedef MediaSelectFn = Future<List<ChosenMedia>> Function(
  BuildContext context, {
  FlutterMediaContentType? mediaType,
  VoidCallback? onLoading,
  int? maxSelections,

  /// By default will show all available
  Set<FileSourceType>? sourceTypes,
  dynamic extraOptions,
});

/// This pops the file picker/cropper and returns a single file.
Future<ChosenMedia?> showImagePicker(
  BuildContext context, {
  FileChanged? onChange,
  FilesSelected? onSelected,
  bool crop = true,
  VoidCallback? onPermissionsInitialized,
  bool preemptPermissions = false,
  bool allowThirdParties = false,
  dynamic extraOptions,
}) async {
  try {
    final imageFn =
        // ignore: deprecated_member_use_from_same_package
        allowThirdParties ? Choosers.flutterFilePicker : Choosers.chooseImages;
    final selection = await showImageListPicker(context,
        onChange: onChange,
        maxSelections: 1,
        crop: crop,
        onSelected: onSelected,
        onPermissionsInitialized: onPermissionsInitialized,
        imageSelector: imageFn,
        preemptPermission: preemptPermissions,
        extraOptions: extraOptions);
    return selection?.firstOrNull();
  } catch (e, stack) {
    _log.severe("Image picker error: $e", e, stack);
    return null;
  }
}

Future<List<ChosenMedia>?> showMediaListPicker(
  BuildContext context,
  FlutterMediaContentType mediaType, {
  required VoidCallback onLoading,
  VoidCallback? onPermissionsInitialized,
  FilesSelected? onSelected,
  FileChanged? onChange,
  int maxSelections = 1,
  double? cropAspectRatio,
  bool crop = true,
  bool preemptPermission = true,
  Set<FileSourceType>? sourceTypes,
  MediaSelectFn imageSelector = Choosers.flutterFilePicker,
  dynamic extraOptions,
}) async {
  final permission = await checkAndRequestPermissions(
    context,
    title: (intl) => intl!.permissionRequestPhotos,
    permission: Permission.photos,
    preemptPermission: preemptPermission,
  );
  switch (permission) {
    case SunnyPermissionStatus.rejected:
      SunnyHud.notification(
          context, "You haven't granted access to your photos.",
          isDismissible: true, onTap: (_, en) {
        openAppSettings();
        en?.dismiss();
      });
      break;
    case SunnyPermissionStatus.granted:
      try {
        onPermissionsInitialized?.call();
        var start = DateTime.now();
        final files = await imageSelector(
          context,
          maxSelections: maxSelections,
          onLoading: () {},
          sourceTypes: sourceTypes,
          mediaType: mediaType,
          extraOptions: extraOptions,
        );
        var end = DateTime.now();

        if (crop == true && files.length == 1 && files.first.file != null) {
          final imageFile = files.first.file;
          onChange?.call(imageFile, false);
          final cropPt = await ((infoX.isIOS || infoX.isAndroid)
              ? Croppers.cropNative(context, imageFile!,
                  cropAspectRatio: cropAspectRatio)
              : Croppers.cropFallback(context, imageFile!,
                  cropAspectRatio: cropAspectRatio));
          // Upload new image
          onChange?.call(cropPt, true);
          return [ChosenMedia.ofFile(cropPt)];
        }
        _log.info("Took ${end.difference(start)} to pick");

        onSelected?.call(files);
        return files;
      } on NoImagesSelectedException {
        return [];
      }
    case SunnyPermissionStatus.later:
      return [];
  }
  return null;
}

Future<List<ChosenMedia>?> showImageListPicker(
  BuildContext context, {
  VoidCallback? onPermissionsInitialized,
  FilesSelected? onSelected,
  FileChanged? onChange,
  int maxSelections = 1,
  bool crop = true,
  bool preemptPermission = true,
  // ignore: deprecated_member_use_from_same_package
  MediaSelectFn imageSelector = Choosers.chooseImages,
  dynamic extraOptions,
}) async {
  final permission = await checkAndRequestPermissions(
    context,
    title: (intl) => intl!.permissionRequestPhotos,
    permission: Permission.photos,
    preemptPermission: preemptPermission,
  );

  switch (permission) {
    case SunnyPermissionStatus.rejected:
      SunnyHud.error(context, "You haven't granted access to your photos.");
      break;
    case SunnyPermissionStatus.granted:
      try {
        onPermissionsInitialized?.call();
        final images = await imageSelector(context,
            maxSelections: maxSelections, extraOptions: extraOptions);
        onSelected?.call(images);
        if (crop == true && images.length == 1 && images.first.file != null) {
          final imageFile = images.first.file;
          onChange?.call(imageFile, false);
          final cropPt = await ((infoX.isIOS || infoX.isAndroid)
              ? Croppers.cropNative(context, imageFile!)
              : Croppers.cropFallback(context, imageFile!));
          // Upload new image
          onChange?.call(cropPt, true);
          return [ChosenMedia.ofFile(cropPt)];
        } else {
          return images;
        }
      } on NoImagesSelectedException {
        return [];
      }
    case SunnyPermissionStatus.later:
      return [];
  }
  return null;
}
