import 'package:flutter/material.dart';
import 'package:sunny_dart/helpers/lists.dart';
import 'package:sunny_service_stubs/options.dart';

class SelectOptionConfig<K, V> implements MetaFieldConfig {
  final bool isDisabled;
  final bool? isRequired;

  /// Provides a baseline filter on any results.
  final Predicate? resultFilter;
  final Widget? prefixIcon;
  final IKeyedOptionsHandler<K, V>? handler;
  final TypeaheadHandler<K, V>? viewHandler;

  const SelectOptionConfig({
    this.prefixIcon,
    this.handler,
    this.viewHandler,
    this.resultFilter,
    this.isDisabled = false,
    this.isRequired,
  });

  @override
  MetaFieldConfig merge(MetaFieldConfig other) {
    SelectOptionConfig<K, V>? otherAsSelect =
        other is SelectOptionConfig<K, V> ? other : null;

    return copyWith(
      isDisabled: other.isDisabled! || isDisabled,
      isRequired: other.isRequired! || isRequired!,
      resultFilter: otherAsSelect?.resultFilter,
      prefixIcon: otherAsSelect?.prefixIcon,
      handler: otherAsSelect?.handler?.cast(),
      viewHandler: otherAsSelect?.viewHandler,
    );
  }

  SelectOptionConfig<K, V> copyWith({
    bool? isDisabled,
    bool? isRequired,
    Predicate? resultFilter,
    Widget? prefixIcon,
    IKeyedOptionsHandler<K, V>? handler,
    TypeaheadHandler<K, V>? viewHandler,
  }) {
    return SelectOptionConfig<K, V>(
      isDisabled: isDisabled ?? this.isDisabled,
      isRequired: isRequired ?? this.isRequired,
      resultFilter: resultFilter ?? this.resultFilter,
      prefixIcon: prefixIcon ?? this.prefixIcon,
      handler: handler ?? this.handler,
      viewHandler: viewHandler ?? this.viewHandler,
    );
  }
}

/// Configuration for text areas
class TextAreaConfig implements MetaFieldConfig {
  final int? minLines;
  final int? maxLines;
  final bool? expands;
  final bool? autocorrect;
  final bool? isDisabled;
  final bool? isRequired;

  const TextAreaConfig({
    this.minLines,
    this.maxLines,
    this.expands,
    this.isDisabled,
    this.isRequired,
    this.autocorrect,
  });

  @override
  TextAreaConfig merge(MetaFieldConfig other) {
    return other is TextAreaConfig
        ? this.copy(
            minLines: other.minLines,
            maxLines: other.maxLines,
            isDisabled: other.isDisabled,
            isRequired: other.isRequired,
            expands: other.expands,
            autocorrect: other.autocorrect,
          )
        : this.copy(isDisabled: other.isDisabled, isRequired: other.isRequired);
  }

  TextAreaConfig copy({
    int? minLines,
    int? maxLines,
    bool? expands,
    bool? autocorrect,
    bool? isDisabled,
    bool? isRequired,
  }) {
    return TextAreaConfig(
      minLines: minLines ?? this.minLines,
      maxLines: maxLines ?? this.maxLines,
      expands: expands ?? this.expands,
      autocorrect: autocorrect ?? this.autocorrect,
      isDisabled: isDisabled ?? this.isDisabled,
      isRequired: isRequired ?? this.isRequired,
    );
  }
}

/// Configuration for text areas
class TextConfig implements MetaFieldConfig {
  final bool? autocorrect;
  final TextInputType? textInputType;
  final TextCapitalization? textCapitalization;
  final bool? isDisabled;
  final bool? isRequired;

  @override
  TextConfig merge(MetaFieldConfig? other) {
    if (other == null) return this;
    if (other is TextConfig) {
      return this.copy(
        autocorrect: other.autocorrect,
        textInputType: other.textInputType,
        textCapitalization: other.textCapitalization,
        isDisabled: other.isDisabled,
        isRequired: other.isRequired,
      );
    } else {
      return this.copy(isDisabled: other.isDisabled, isRequired: isRequired);
    }
  }

  const TextConfig({
    this.autocorrect,
    this.textInputType,
    this.textCapitalization,
    this.isDisabled,
    this.isRequired,
  });

  TextConfig copy({
    bool? autocorrect,
    TextInputType? textInputType,
    TextCapitalization? textCapitalization,
    bool? isDisabled,
    bool? isRequired,
  }) {
    return TextConfig(
      autocorrect: autocorrect ?? this.autocorrect,
      textInputType: textInputType ?? this.textInputType,
      textCapitalization: textCapitalization ?? this.textCapitalization,
      isDisabled: isDisabled ?? this.isDisabled,
      isRequired: isRequired ?? this.isRequired,
    );
  }
}

extension MetaFieldConfigExtensions on MetaFieldConfig {
  bool get isEnabled => isDisabled != true;
}
