import 'dart:async';

import 'package:flutter/widgets.dart';
import 'package:meta_forms/meta_forms.dart';
import 'package:provider/provider.dart';
import 'package:sunny_dart/sunny_dart.dart';
import 'package:sunny_forms/meta/meta_form.dart';
import 'package:sunny_forms/meta/meta_form_service.dart';
import 'package:sunny_forms/meta/meta_property_handler.dart';
import 'package:sunny_forms/meta/props/hidden_property_handler.dart';
import 'package:sunny_forms/sunny_form_controller.dart';
import 'package:sunny_service_stubs/options.dart';

/// Tracks metadata that helps render meta forms.  Things like labels, placeholders,
/// types, options, ignored fields, etc.
class MetaFormContext with LoggingMixin implements IMetaFormContext {
  final BuildContext context;
  final JsonPath? embedPath;
  final MBaseModel? subject;
  final SafeCompleter<IMSchemaDefinition> _loaded;

  Future<IMSchemaDefinition> get loaded => _loaded.future;

  final Map<String, dynamic> scope;
  final MSchemaRef ref;
  final MetadataOverrides overrides;
  final List<IMetaPropertyHandler> extraHandlers;
  late FutureOr<IMSchemaDefinition> _schema;

  MetaFormResult? handlers;

  IMSchemaDefinition get schema => _schema.resolve()!;

  bool get isLoaded => _schema is Future;

  bool get isEmbed => embedPath?.isNotRoot == true;

  Iterable<JsonPath> get handledPaths =>
      checkLoaded(() => handlers?.handled.keys ?? {});

  set schema(IMSchemaDefinition schema) {
    try {
      bool isLoaded = _schema is Future;
      this._schema = schema;
      this._properties = mapOf(schema.properties.embedded(embedPath)!,
          keyOf: (prop) => prop.jsonPath);
      if (isLoaded) return;

//    _properties = mapOf(
//      schema.properties.embedded(embedPath),
//      keyOf: (prop) => prop.jsonPath,
//    );

      handlers = metaFormService.process(
        schema,
        basePath: embedPath,
        adhocHandlers: [
          ...extraHandlers,
          if (overrides.ignoredPaths.isNotNullOrEmpty == true)
            HiddenPropertyHandler(overrides.ignoredPaths.orEmptySet()),
        ],
      );
      _loaded.complete(schema);
    } catch (e, stack) {
      log.severe("Error initializing form: $e", e, stack);
    }
  }

  late FutureOr<Map<JsonPath, IMSchemaProperty>> _properties;

  Map<JsonPath, IMSchemaProperty>? get properties => _properties.resolve();

  MetaFormContext._(this.context,
      {required this.subject,
      this.handlers,
      required this.ref,
      Map<String, dynamic>? scope,
      MetadataOverrides? overrides,
      List<IMetaPropertyHandler>? extraHandlers,
      FutureOr<IMSchemaDefinition>? schema,
      required this.embedPath})
      : _loaded = SafeCompleter(),
        extraHandlers = [
          ...?extraHandlers,
          if (overrides?.ignoredPaths != null)
            HiddenPropertyHandler(overrides!.ignoredPaths!),
        ],
        overrides = overrides ?? metaFormService.getOverrides(ref),
        scope = scope ?? {} {
    _schema = schema ?? ((ref.getMSchemaOrNull(context)));
    _schema.thenOr((resolved) {
      this.schema = resolved;
    });
  }

  factory MetaFormContext.ofModelForm(BuildContext context,
      {required MModelForm controller,
      required FutureOr<IMEntityDefinition> definition,
      Map<String, dynamic>? scope,
      MetadataOverrides? overrides,
      List<MetaPropertyHandler>? extraHandlers,
      JsonPath? embedPath}) {
    final delegate = controller;
    return MetaFormContext._(
      context,
      ref: delegate.mtype,
      subject: delegate.mmodel,
      schema: definition,
      extraHandlers: extraHandlers,
      overrides: overrides,
      scope: scope,
      embedPath: embedPath,
    );
  }

  MetaFormContext.empty(BuildContext context, MSchemaRef emptyRef)
      : this._(
          context,
          subject: null,
          scope: {},
          ref: emptyRef,
          embedPath: null,
        );

  MetaFormContext(
    BuildContext context, {
    required MBaseModel? subject,
    Map<String, dynamic>? scope,
    required MSchemaRef ref,
    IMSchemaDefinition? schema,
//    FormController controller,
    MetadataOverrides? overrides,
    JsonPath? embedPath,
    List<MetaPropertyHandler>? extraHandlers,
  }) : this._(context,
            subject: subject,
            scope: scope,
            ref: ref,
            schema: schema,
            overrides: overrides,
            extraHandlers: extraHandlers,
            embedPath: embedPath);

  MetaFormContext.embedded(
    BuildContext context, {
    required MModel subject,
    Map<String, dynamic>? scope,
    required MSchemaRef ref,
    IMSchemaDefinition? schema,
    MetadataOverrides? overrides,
    List<MetaPropertyHandler>? extraHandlers,
    required JsonPath embedPath,
  }) : this._(context,
            ref: ref,
            subject: subject,
            scope: scope,
            schema: schema,
            overrides: overrides,
            extraHandlers: extraHandlers,
            embedPath: embedPath);

  T checkLoaded<T>(T invoke()) {
    if (_loaded.isNotComplete) throw "Meta form not loaded";
    return invoke();
  }

  IMetaFormFieldContext prop(IMSchemaProperty prop) =>
      checkLoaded(() => MetaFormFieldContext(
            form: this,
            renderer: handlers!.handled[prop.jsonPath],
            formConfig: this.overrides.formConfig![prop.jsonPath] ?? [],
            fullPath: prop.jsonPath,
            property: prop,
          ));

  MetaFormFieldContext relPath(JsonPath fieldPath) =>
      this.path(this.embedPath!, fieldPath);

  MetaFormFieldContext path(JsonPath basePath, JsonPath path) {
    final fullPath = (basePath + path);
//    nullPointer("No property found at $fullPath - found ${properties.keys}");
    return checkLoaded(() {
      var propFieldConfig = this.overrides.formConfig![fullPath] ?? [];
      return MetaFormFieldContext(
        form: this,
        fullPath: fullPath,
        formConfig: propFieldConfig,
        renderer: handlers!.get(embedPath, fullPath),
        property: properties![fullPath] ??
            IMSchemaProperty.of(
              type: "stringSchema",
              path: "$fullPath",
              uri: Definitions.titleField.uri.toUri(),
              baseCode: Definitions.titleField.baseCode,
              isRequired:
                  propFieldConfig.any((element) => element.isRequired == true),
              label: "$fullPath",
            ),
      );
    });
  }

  MetaFormContext withOverrides(MetadataOverrides? overrides) {
    if (overrides == null) {
      return this;
    } else {
      return copyWith(overrides: (this.overrides) + overrides);
    }
  }

  MetaFormContext copyWith({
    JsonPath? embedPath,
    MModel? subject,
    MetaFormResult? handlers,
    MetaPropertyHandler? renderer,
    IMSchemaDefinition? schema,
    Map<String, dynamic>? scope,
    MetadataOverrides? overrides,
    FormController? controller,
    List<MetaPropertyHandler>? extraHandlers,
  }) {
    return MetaFormContext._(
      context,
      ref: ref,
      subject: subject ?? this.subject,
      handlers: handlers ?? this.handlers,
      schema: schema ?? this._schema,
      embedPath: embedPath ?? this.embedPath,
      scope: scope ?? this.scope,
      overrides: overrides ?? this.overrides,
      extraHandlers: extraHandlers ?? this.extraHandlers,
    );
  }

  HandledPaths? getHandler(JsonPath<dynamic> jsonPath) => checkLoaded(() {
        return handlers!.handled[jsonPath];
      });
}

class MetaFormFieldContext implements IMetaFormFieldContext {
  final MetaFormContext form;
  final JsonPath fullPath;
  IMSchemaProperty? property;
  final HandledPaths? renderer;
  final Map<String, dynamic> data;
  final List<MetaFieldConfig> _formConfig;

  bool? get isRequired => property?.isRequired;

  MetaFormFieldContext({
    required this.form,
    Map<String, dynamic>? data,
    this.renderer,
    required List<MetaFieldConfig> formConfig,
    required this.fullPath,
    required IMSchemaProperty this.property,
  })  : _formConfig = formConfig,
        data = data ?? {};

  MetaFormFieldContext.noop(BuildContext context, MSchemaRef emptyRef)
      : form = MetaFormContext.empty(context, emptyRef),
        fullPath = JsonPath.root(),
        renderer = null,
        _formConfig = const [],
        data = const {},
        property = null;

  T? formConfig<T extends MetaFieldConfig>() {
    final configs = _formConfig.whereType<T>();
    if (configs.isEmpty) {
      return null;
    } else if (configs.length == 1) {
      return configs.first;
    } else {
      final reduced =
          configs.cast<MetaFieldConfig>().reduce((a, b) => a.merge(b));
      return reduced as T;
    }
  }

  MetadataOverrides get overrides => form.overrides;

  List<IMetaPropertyHandler> get extraHandlers => form.extraHandlers;

  String? resolveLabel(
      {String? labelOverride, IMSchemaProperty? prop, String? fallbackLabel}) {
    prop ??= property?.resolveOrNull();
    final label = labelOverride ??
        overrides.resolveLabel(fullPath) ??
        prop?.label ??
        fallbackLabel;
    return label;
  }

  String? resolvePlaceholder({String? placeholderOverride}) {
    final placeholder =
        placeholderOverride ?? overrides.resolvePlaceholder(fullPath);
    return placeholder;
  }
}

extension FormControllerToContext on FormController {
  MetaFormContext toMetaFormContext(BuildContext context,
      {Map<String, dynamic>? scope,
      MetadataOverrides? overrides,
      List<MetaPropertyHandler>? extraHandlers,
      FutureOr<IMEntityDefinition>? definition}) {
    if (this is MModelForm) {
      final form = this as MModelForm;

      definition ??= context.mschemaService.getEntityDefinition(form.mtype);
      return MetaFormContext.ofModelForm(
        context,
        controller: form,
        definition: definition,
        scope: scope,
        extraHandlers: extraHandlers,
        overrides: overrides,
      );
    }
    throw UnimplementedError("Not implemented");
  }
}

extension MetaFormBuildContext on BuildContext {
  MetaFormContext get formMetadata {
    return Provider.of(this);
  }

  MetaFormContext? get formMetadataOrNull {
    try {
      return Provider.of(this);
    } on ProviderNotFoundException {
      return null;
    }
  }
}

// MetaFormConstructor() {
//   /// This constructor expects form metadata to be the source for the field rendering, as well
//   /// as a [SunnyForm] to be present.
//   factory SunnyFormField.meta(
//       BuildContext context, {
//   Key key,
//   // Optional
//   String label,
//
//   // Optional
//   String fallbackLabel,
//
//   // Optional
//   String placeholder,
//   dynamic basePath,
//   // If this isn't provided, it will be lookup up using a Provider
//   IMetaFormContext formMetadata,
//   @required dynamic attribute,
//   bool captureNestedPaths,
//   bool isReadOnly = false,
//   bool isRequired,
//   AttributeAccepter<T> onReadOnlyTapped,
//   FormValueChanged<T> onChange,
//   }) {
//     var _basePath = JsonPath.of(basePath);
//     var _attribute = JsonPath.of(attribute);
//     final fullPath = (_basePath + _attribute).cast<T>();
//     final meta = formMetadata ?? context.formMetadata;
//     final metaField = meta.path(_basePath, _attribute);
//     return SunnyFormField<T>(
//       key: key ?? Key("sunnyForm-field-${fullPath.toKey()}"),
//       attribute: fullPath,
//       builder: (context, state) {
//         return MetaFormControl<T>.ofJsonPath(
//           path: fullPath,
//           metaForm: meta,
//           fieldState: state,
//         );
//       },
//       captureNestedPaths: captureNestedPaths == true,
//       isReadOnly: isReadOnly,
//       isRequired: isRequired ?? metaField?.isRequired ?? false,
//       onReadOnlyTapped: onReadOnlyTapped,
//       onChange: onChange,
//       label: metaField.resolveLabel(
//           labelOverride: label, fallbackLabel: fallbackLabel),
//       placeholder:
//       metaField.resolvePlaceholder(placeholderOverride: placeholder),
//     );
//   }
// }
