import 'dart:async';

import 'package:flutter/material.dart';
import 'package:flutter/widgets.dart';
import 'package:meta_forms/meta_forms.dart';
import 'package:sunny_core_widgets/sunny_core_widgets.dart';
import 'package:sunny_dart/sunny_dart.dart';
import 'package:sunny_forms/builders/future.dart';
import 'package:sunny_forms/meta/meta_form_context.dart';
import 'package:sunny_forms/meta/meta_property_handler.dart';
import 'package:sunny_forms/sunny_form_controller.dart';
import 'package:sunny_forms/sunny_form_field.dart';
import 'package:sunny_service_stubs/options.dart';

import 'meta_form_service.dart';

class MetaFormControl<M> extends StatelessWidget with LoggingMixin {
  final MetaFormFieldContext metaFormField;
  final SunnyFormFieldState<M?> fieldState;

  final JsonPath basePath;
  final JsonPath fieldPath;
  final JsonPath fullPath;

  final Key? widgetKey;

  final String? label;
  final String? placeholder;

  final bool checkAll;

  MetaFormControl._({
    this.widgetKey,
    required this.fieldState,
    required JsonPath fieldPath,
    required this.metaFormField,
    this.label,
    this.checkAll = true,
    this.placeholder,
    required JsonPath basePath,
  })  : basePath = basePath.self,
        fieldPath = fieldPath.self,
        fullPath = basePath + fieldPath,
        super(
            key: Key(
                "sunnyForm-metaFormField-${(basePath + fieldPath).toKey()}"));

  factory MetaFormControl.ofMSchema(
    BuildContext context, {
    Key? widgetKey,
    required JsonPath fieldPath,
    required MSchemaForm form,
    Map<String, dynamic>? scope,
    MetadataOverrides? overrides,
    required SunnyFormFieldState<M> fieldState,
    List<MetaPropertyHandler>? extraHandlers,
    String? label,
    bool checkAll = true,
    String? placeholder,
    JsonPath? basePath,
  }) {
    final metaProp = form.definition!.prop(fieldPath);
    return MetaFormControl._(
      widgetKey: widgetKey,
      fieldPath: fieldPath,
      checkAll: checkAll,
      fieldState: fieldState,
      metaFormField: MetaFormFieldContext(
          formConfig: overrides?.formConfig?.orEmpty()[fieldPath] ?? [],
          form: MetaFormContext.ofModelForm(
            context,
            controller: form,
            embedPath: basePath,
            scope: scope,
            overrides: overrides,
            extraHandlers: extraHandlers,
            definition: form.definition!,
          ),
          property: metaProp!,
          fullPath: (basePath! + fieldPath)),
      basePath: basePath.self,
      label: label,
      placeholder: placeholder,
    );
  }

  /// This doesn't allow embedding, just to ease on confusion.  If you want an embedded field
  /// use the constructor below [MetaFormControl.embedded]
  MetaFormControl.ofJsonPath({
    Key? widgetKey,
    required JsonPath path,
    bool checkAll = true,
    required SunnyFormFieldState<M?> fieldState,
    required MetaFormContext metaForm,
  }) : this._(
          widgetKey: widgetKey,
          basePath: JsonPath.Root,
          fieldPath: path,
          checkAll: checkAll,
          metaFormField: metaForm.path(JsonPath.Root, path),
          fieldState: fieldState,
        );

  MetaFormControl.embedded({
    Key? widgetKey,
    required SunnyFormFieldState<M> fieldState,
    required JsonPath basePath,
    required JsonPath path,
    required MetaFormContext metaForm,
    String? label,
    bool checkAll = true,
    String? placeholder,
  }) : this._(
          widgetKey: widgetKey,
          fieldState: fieldState,
          basePath: basePath.verifyNotRoot,
          fieldPath: path,
          metaFormField: metaForm.path(basePath, path),
          label: label,
          checkAll: checkAll,
          placeholder: placeholder,
        );

  FutureOr<IMSchemaProperty>? get _prop => metaFormField.property;

  @override
  Widget build(BuildContext context) {
    return future<IMSchemaProperty>(context,
        waitFor: _prop,
        initialData: _prop.resolveOrNull() as IMSchemaProperty?,
        allowNull: true, builder: (context, prop) {
      if (prop == null) return emptyBox;
      final renderer = metaFormField.renderer ??
          () {
            return metaFormService.processSingle(
              metaFormField.form.schema,
              basePath: this.basePath,
              singlePath: this.fieldPath,
            );
          }();
      return column([
        ...renderer.render(context, metaFormField.form, fieldState,
            key: widgetKey),
      ]);
    });
  }
}

class MetaFormSection {
  final String? label;
  final String? key;
  final Set<JsonPath> properties;

  const MetaFormSection._({this.label, this.key, Set<JsonPath>? properties})
      : properties = properties ?? const {};

  MetaFormSection({String? label, String? key, required Set properties})
      : this._(
            label: label,
            key: key,
            properties: {...properties.map((prop) => JsonPath.of(prop))});

  bool get isNotEmpty => properties.isNotEmpty;
  bool get isEmpty => properties.isEmpty;

  operator +(JsonPath property) {
    final properties = Set.from(this.properties);
    properties.add(property);
    return MetaFormSection(label: label, properties: properties, key: key);
  }

  filtered(Set<JsonPath> against) {
    return MetaFormSection(key: key, label: label, properties: {
      ...properties.where((path) => against.contains(path)),
    });
  }
}
