import 'package:flutter/material.dart';
import 'package:flutter/widgets.dart';
import 'package:meta_forms/meta_forms.dart';
import 'package:sunny_dart/sunny_dart.dart';
import 'package:sunny_forms/sunny_form_field.dart';
import 'package:sunny_service_stubs/options.dart';

/// Inspects a schema and looks for specialized properties for which we want to build a special UI or component.  Some
/// fields might be grouped together, for example
/// [T] The type of data this handler should encounter.
abstract class MetaPropertyHandler<T> implements IMetaPropertyHandler<T> {
  /// Examines a form, and returns a list of properties that this handler is capable of dealing with, grouped together
  /// in [HandledPaths].  Each of these field clusters will be passed back to the handler during rendering
  Iterable<HandledPaths> acceptProperties(
      JsonPath? basePath, IMSchemaDefinition definition, Map<JsonPath, IMSchemaProperty> available);

  /// Renders a widget (or null) for a provided [HandledPaths] (see [acceptProperties])
  Iterable<Widget> renderProperty(
      BuildContext context, IMetaFormContext metaForm, HandledPaths paths, SunnyFormFieldState<T> state,
      {Key? key});

  SunnyFormFieldState<T> castState<X>(SunnyFormFieldState state) {
    return state.cast<T>();
  }
}

/// A list of json paths - sometimes rendering will handle more than one property at a time
class HandledPaths {
  /// The path to this handler - by default, it's the first path that's being handled
  final JsonPath path;

  /// A base path to be applied
  final JsonPath basePath;

  /// The paths being handled together
  final Map<JsonPath, IMSchemaProperty?> paths;

  final dynamic data;

  /// The handler that is processing these attributes
  final MetaPropertyHandler _handler;

  bool get isEmbed => basePath.isNotRoot;

  IMSchemaProperty? get property {
    assert(paths.length == 1, "Multiple properties - can only use this getter for single paths");
    return paths.entries.first.value;
  }

  JsonPath<T> fullPath<T>() {
    assert(paths.length == 1, "Multiple properties - can only use this getter for single paths");
    return (basePath + path).cast<T>();
  }

  Iterable<Widget> render(BuildContext context, IMetaFormContext metaForm, SunnyFormFieldState state, {Key? key}) =>
      _handler.renderProperty(context, metaForm, this, _handler.castState(state), key: key);

  HandledPaths.ofProperty(IMSchemaProperty prop, this.basePath, this._handler, {this.data = const {}})
      : path = prop.jsonPath,
        paths = {prop.jsonPath: prop};

  HandledPaths.ofProperties(Iterable<IMSchemaProperty?> properties, this.basePath, this._handler, {this.data = const {}})
      : paths = mapOf(properties, keyOf: (p) => p!.jsonPath),
        path = properties.first!.jsonPath;
}
