import 'package:flutter/material.dart';
import 'package:sunny_core_widgets/sunny_core_widgets.dart';
import 'package:sunny_dart/sunny_dart.dart';
import 'package:sunny_forms/form_fields/select_field.dart';
import 'package:sunny_forms/icon/image_widgets.dart';
import 'package:sunny_forms/widget_extensions.dart';
import 'package:sunny_sdk_core/mverse.dart';
import 'package:sunny_service_stubs/sunny_service_stubs.dart';

typedef SuggestionElement<K, T> = dynamic Function(
    KeyedOption<K, T> suggestion);

TypeaheadService get optionViewService => sunny.get();

/// Service that contains all registered typeahead handlers.  Works with [OptionService]
class TypeaheadService {
  final Map<String?, TypeaheadHandler> _typeaheads = {};

  TypeaheadHandler<K, T> getTypeaheadRenderer<K, T>(Object? ref) {
    final key = ref.baseCode;
    final renderer = _typeaheads[key] as TypeaheadHandler<K, T>? ??
        defaultOptionViewHandler(showExpand: true, noOptionsLabel: "N/A");
    return renderer;
  }

  void operator []=(Object key, TypeaheadHandler handler) {
    _typeaheads[key.baseCode] = handler;
  }

  void registerTypeaheadRenderer<K, T>(
      Object ref, TypeaheadHandler<K, T> renderer) {
    final key = ref.baseCode;
    _typeaheads[key] = renderer;
  }

  TypeaheadService();
}

extension TypeaheadFocusModeExt on TypeaheadFocusMode? {
  bool get isShowAll {
    return this == TypeaheadFocusMode.showAll ||
        this == TypeaheadFocusMode.showAllAndClear;
  }
}

TypeaheadOptionsAndHandler<K, T> defaultOptionViewHandler<K, T>(
    {Icon? prefixIcon,
    String? placeholder,
    String noOptionsLabel = "N/A",
    OptionValueType? valueType,
    TypeaheadFocusMode? focusMode,
    bool showExpand = false,
    RenderSuggestionTile<K, T>? renderSuggestionTile}) {
  return DefaultOptionViewHandler<K, T>(
    prefixIcon: prefixIcon,
    placeholder: placeholder,
    valueType: valueType ?? OptionValueType.key,
    noOptionsLabel: noOptionsLabel,
    renderSuggestionTile: renderSuggestionTile,
    focusMode: focusMode ?? TypeaheadFocusMode.showAll,
    showExpand: showExpand,
  );
}

class DefaultOptionViewHandler<K, T> extends TypeaheadOptionsAndHandler<K, T>
    with DefaultOptionViewMixin2<K, T> {
  final Icon? prefixIcon;
  final String? placeholder;
  final String noOptionsLabel;
  final TypeaheadFocusMode focusMode;
  final bool showExpand;
  final OptionValueType valueType;

  final RenderSuggestionTile<K, T>? _renderSuggestionTile;

  const DefaultOptionViewHandler(
      {this.prefixIcon,
      required this.valueType,
      this.focusMode = TypeaheadFocusMode.showAll,
      this.placeholder,
      required this.noOptionsLabel,
      required this.showExpand,
      RenderSuggestionTile<K, T>? renderSuggestionTile})
      : _renderSuggestionTile = renderSuggestionTile,
        super();

  @override
  Widget renderSuggestionTile(
      BuildContext context, KeyedOption<K, T> suggestion,
      {required bool isSelected,
      required SelectOption<K, T> selectOption,
      VoidCallback? onTap}) {
    final Widget Function(BuildContext context, KeyedOption<K, T> value,
            {required bool isSelected,
            required void Function(KeyedOption<K, T> option) selectOption,
            void Function()? onTap}) _renderSuggestionTile =
        this._renderSuggestionTile ?? super.renderSuggestionTile;
    final tile = _renderSuggestionTile(
      context,
      suggestion,
      isSelected: isSelected,
      selectOption: selectOption,
      onTap: onTap,
    );
    return this.wrapSuggestionTile(tile);
  }
}

mixin DefaultOptionViewMixin2<K, T> on TypeaheadHandler<K, T> {
  /// Gets the selected text that shows up in the
  @override
  String? getSelection(KeyedOption<K, T>? selectedOption) =>
      selectedOption?.selection;

  /// Renders the suffix for the selected item in the form control
  @override
  Widget? renderSelectedItemSuffix(
          BuildContext context, KeyedOption<K, T> selected,
          {required SelectOption<K, T> selectOption}) =>
      null;

  /// Renders the suffix for an item that shows up in the list of options.  If this method returns null, then
  /// the leading spot will not be rendered

  Widget? renderSuggestionTileTrailing(
    BuildContext context,
    KeyedOption<K, T> option,

    /// Whether the item we're building the trailing for is currently selected
    bool isSelected,

    /// A callback that can be used to change the bound form control selection
    SelectOption<K, T> selectOption,
  ) {
    return null;
  }

  @override
  Widget wrapSuggestionTile(Widget tile) => tile;

  /// Renders the suggested tile in the list of options.  If you override this method, you should make sure to invoke
  /// [renderSuggestionTileTrailing]
  @override
  Widget renderSuggestionTile(
      BuildContext context, KeyedOption<K, T> suggestion,
      {required bool isSelected,

      /// A callback that allows you to change the selected option
      required SelectOption<K, T> selectOption,

      /// A provided onTap handler for the tile.
      VoidCallback? onTap}) {
    final isAdhocCreator = suggestion is KeyedAdhocOption<K, T> &&
        suggestion.hasAdhocCreator &&
        suggestion.label.isNotNullOrBlank;
    Widget? subtitle() {
      if (isAdhocCreator) {
        return Text("Tap to add this record");
      } else {
        return suggestion.subtitle.toColumn();
      }
    }

    return InputWidgets.buildTypeaheadOption(
      onTap: onTap,
      leading: imageWidgetOf(context, suggestion.icon, size: 30, circle: false),
      title: Text(suggestion.label),
      subtitle: subtitle(),
      trailing: isAdhocCreator
          ? context.icons.add.widget
          : renderSuggestionTileTrailing(
              context,
              suggestion,
              isSelected,
              selectOption,
            ),
    );
  }
}

/// Makes it easy to produce a formatted suggestion tile by proving a [SuggestionElement] callback for the
/// title and icon
RenderSuggestionTile<K, T> renderSimpleTypeaheadOption<K, T>(
    {required SuggestionElement<K, T> title,
    SuggestionElement<K, T>? icon,
    bool isAdhocCreator = false}) {
  return (BuildContext context, suggestion,
      {bool? isSelected, SelectOption<K, T>? selectOption, onTap}) {
    final isAdhocCreator = suggestion is KeyedAdhocOption<K, T> &&
        suggestion.hasAdhocCreator &&
        suggestion.label.isNotNullOrBlank;
    Widget? subtitle() {
      if (isAdhocCreator) {
        return Text("Tap to add this record");
      } else {
        return suggestion.subtitle.toColumn();
      }
    }

    return TypeaheadOption(
      child: ListTile(
        dense: true,
        onTap: onTap,
        leading: imageWidgetOf(context, icon?.call(suggestion),
            size: 30, circle: false),
        title: TileTitle.of(title(suggestion)),
        subtitle: subtitle(),
      ),
    );
  };
}

extension TypeaheadHandlerExtensions<K, T> on TypeaheadHandler<K, T> {
  TypeaheadOptions? get options {
    if (this is TypeaheadOptions) {
      return this as TypeaheadOptions;
    }
    return null;
  }
}

extension BuildContextTypeaheadService on BuildContext {
  TypeaheadService get typeaheadService {
    return sunny.get(context: this);
  }
}

typedef RenderSuggestionTile<K, T> = Widget Function(
    BuildContext context, KeyedOption<K, T> suggestion,
    {bool? isSelected, SelectOption<K, T>? selectOption, VoidCallback? onTap});
