/*
 * sunny_platform_widgets
 * Copyright (c) 2018 Lance Johnstone. All rights reserved.
 * See LICENSE for distribution and usage details.
 */

import 'dart:ui';

import 'package:dartxx/dartxx.dart';
import 'package:flutter/cupertino.dart';
import 'package:flutter/material.dart';
import 'package:sunny_platform_widgets/sunny_platform_widgets.dart';
import 'package:sunny_platform_widgets/sunny_platform_widgets.dart' as pf;
import 'package:sunny_core_widgets/sunny_core_widgets.dart';
import 'package:sunny_forms/builders/layout.dart';
import 'package:sunny_forms/platform/platform_bottom_sheet_action.dart';
import 'package:sunny_forms/sunny_colors.dart';

import 'platform_stateful_widget_base.dart';

abstract class _BaseData {
  _BaseData({
    this.widgetKey,
    this.content,
    this.title,
    this.cancelButton,
  });

  final Key? widgetKey;
  final Widget? content;
  final Widget? title;
  final Widget? cancelButton;
}

class MaterialBottomSheetData extends _BaseData {
  MaterialBottomSheetData({
    Key? widgetKey,
    Widget? content,
    Widget? title,
    this.onClosing,
    this.builder,
    this.enableDrag,
    this.actions,
    this.elevation,
  }) : super(widgetKey: widgetKey, content: content, title: title);

  final Iterable<Widget>? actions;

  /// Called when the bottom sheet begins to close.
  ///
  /// A bottom sheet might be prevented from closing (e.g., by user
  /// interaction) even after this callback is called. For this reason, this
  /// callback might be call multiple times for a given bottom sheet.
  final VoidCallback? onClosing;

  /// A builder for the contents of the sheet.
  ///
  /// The bottom sheet will wrap the widget produced by this builder in a
  /// [Material] widget.
  final WidgetBuilder? builder;

  /// If true, the bottom sheet can dragged up and down and dismissed by swiping
  /// downwards.
  ///
  /// Default is true.
  final bool? enableDrag;

  /// The z-coordinate at which to place this material relative to its parent.
  ///
  /// This controls the size of the shadow below the material.
  ///
  /// Defaults to 0. The value is non-negative.
  final double? elevation;
}

class CupertinoActionSheetData extends _BaseData {
  CupertinoActionSheetData(
    this.messageScrollController,
    Widget cancelButton, {
    Key? widgetKey,
    this.actions,
    Widget? message,
    Widget? title,
    this.actionScrollController,
  }) : super(
            widgetKey: widgetKey,
            content: message,
            title: title,
            cancelButton: cancelButton);

  final Iterable<CupertinoActionSheetAction>? actions;

  /// A scroll controller that can be used to control the scrolling of the
  /// [messageBuilder] in the action sheet.
  ///
  /// This attribute is typically not needed, as alert messages should be
  /// short.
  final ScrollController messageScrollController;

  /// A scroll controller that can be used to control the scrolling of the
  /// [actions] in the action sheet.
  ///
  /// This attribute is typically not needed.
  final ScrollController? actionScrollController;
}

class PlatformBottomSheet extends PlatformWidgetBase<Widget, Widget> {
  final Key? widgetKey;
  final Iterable<Widget>? actions;
  final Widget? content;
  final Widget? cancelButton;
  final Widget? title;
  final AnimationController? animationController;

  final PlatformBuilder<MaterialBottomSheetData>? android;
  final PlatformBuilder<CupertinoActionSheetData>? ios;

  PlatformBottomSheet(
      {Key? key,
      this.widgetKey,
      this.actions,
      this.content,
      this.title,
      this.ios,
      this.android,
      this.animationController,
      this.cancelButton,
      extraOptions})
      : super(key: key, extraOptions: extraOptions);

  @override
  Widget createCupertinoWidget(BuildContext context) {
    return _PlatformBottomSheet(this);
  }

  @override
  Widget createMaterialWidget(BuildContext context) {
    return _PlatformBottomSheet(this);
  }
}

class _PlatformBottomSheet
    extends PlatformStatefulWidgetBase<CupertinoActionSheet, BottomSheet> {
  final PlatformBottomSheet _widget;

  _PlatformBottomSheet(this._widget);

  @override
  State<StatefulWidget> createState() {
    return PlatformBottomSheetState();
  }
}

class PlatformBottomSheetState extends PlatformWidgetState<CupertinoActionSheet,
    Widget, _PlatformBottomSheet> with SingleTickerProviderStateMixin {
  PlatformBottomSheet get _widget => widget._widget;

  PlatformBuilder<MaterialBottomSheetData>? get android => _widget.android;

  PlatformBuilder<CupertinoActionSheetData>? get ios => _widget.ios;
  AnimationController? animationController;

  @override
  Widget createMaterialWidget(BuildContext context) {
    MaterialBottomSheetData? data;
    if (android != null) {
      data = android!(context, platform(context));
    }
    return Container(
      margin: SunnyMargins.all,
      color: Colors.transparent,
      alignment: Alignment.bottomCenter,
      child: Container(
        constraints: BoxConstraints(maxWidth: 500),
        decoration:
            BoxDecoration(color: sunnyColors.white, borderRadius: tenPxRadius),
        child: Column(
          mainAxisSize: MainAxisSize.min,
          crossAxisAlignment: CrossAxisAlignment.stretch,
          children: [
            if ((data?.title ?? _widget.title) != null)
              Container(
                padding: SunnyMargins.all,
                alignment: Alignment.center,
                child: data?.title ?? _widget.title,
              ),
            if ((data?.content ?? _widget.content) != null)
              Container(
                padding: SunnyMargins.all,
                alignment: Alignment.center,
                child: data?.content ?? _widget.content,
              ),
            ...(data?.actions ?? _widget.actions ?? <Widget>[]),
            if (platformWidget._widget.cancelButton != null) verticalSpace,
            if (platformWidget._widget.cancelButton != null)
              platformWidget._widget.cancelButton!,
          ],
        ),
      ),
    );
  }

  @override
  CupertinoActionSheet createCupertinoWidget(BuildContext context) {
    CupertinoActionSheetData? data;
    if (ios != null) {
      data = ios!(context, platform(context));
    }

    final Iterable<Widget> base =
        (data?.actions ?? _widget.actions ?? const <Widget>[]);
    final List<Widget> actions = base.map(_creatIosAction).notNull().toList();

    return CupertinoActionSheet(
      key: data?.widgetKey ?? _widget.widgetKey,
      title: data?.title ?? _widget.title,
      message: data?.content ?? _widget.content,
      actions: actions,
      cancelButton: data?.cancelButton ?? _widget.cancelButton,
      actionScrollController: data?.actionScrollController,
      messageScrollController: data?.messageScrollController,
    );
  }

  Widget? _creatIosAction(Widget? item) {
    if (item == null) {
      return null;
    }
    if (item is PlatformBottomSheetAction) {
      return item.createCupertinoWidget(context);
    } else {
      return item;
    }
  }

  @override
  void initState() {
    super.initState();
  }

  @override
  void didChangeDependencies() {
    super.didChangeDependencies();
    animationController ??= isMaterial(context)
        ? _widget.animationController ?? AnimationController(vsync: this)
        : null;
  }

  @override
  void dispose() {
    animationController?.dispose();
    super.dispose();
  }
}

typedef BottomSheetBuilder = Widget Function(BuildContext context);

Future<T?> showPlatformBottomSheet<T>({
  required BuildContext context,
  required BottomSheetBuilder builder,
  extraOptions,
}) {
  return pf.showPlatformModalSheet<T>(
    context: context,
    builder: builder,
    extraOptions: extraOptions,
    material: MaterialModalSheetData(
      backgroundColor: Colors.transparent,
      barrierColor: sunnyColors.barrierColor,
      isDismissible: true,
    ),
    cupertino: CupertinoModalSheetData(
        imageFilter: ImageFilter.blur(
          sigmaX: 0.001,
          sigmaY: 0.001,
        ),
        useRootNavigator: true,
        semanticsDismissible: true),
  );
}
