import 'dart:async';

import 'package:flutter/cupertino.dart';
import 'package:flutter/material.dart';
import 'package:flutter/services.dart';
import 'package:flutter_typeahead/flutter_typeahead.dart';
import 'package:sunny_dart/sunny_dart.dart';
import 'package:sunny_forms/platform/platform_stateful_widget_base.dart';

typedef FutureOr<List<T>> SuggestionsCallback<T>(String pattern);
typedef Widget ItemBuilder<T>(BuildContext context, T itemData);
typedef void SuggestionSelectionCallback<T>(T suggestion);
typedef Widget ErrorBuilder(BuildContext context, Object? error);

class PlatformTypeAhead<T> extends PlatformStatefulWidgetBase<
    CupertinoTypeAheadField<T>, TypeAheadField<T>> {
  final SuggestionsCallback<T> suggestionsCallback;
  final SuggestionSelectionCallback<T> onSuggestionSelected;
  final ItemBuilder<T> itemBuilder;
  final SuggestionsBoxDecoration? suggestionsBoxDecoration;
  final PlatformSuggestionBoxController? suggestionsBoxController;
  final Duration? debounceDuration;
  final WidgetBuilder? loadingBuilder;
  final WidgetBuilder? noItemsFoundBuilder;
  final ErrorBuilder? errorBuilder;
  final AnimationTransitionBuilder? transitionBuilder;
  final Duration? animationDuration;
  final AxisDirection? direction;
  final double animationStart;
  final PlatformTextFieldConfiguration<T>? textFieldConfiguration;
  final double suggestionsBoxVerticalOffset;

  final bool? getImmediateSuggestions;
  final bool? hideOnLoading;
  final bool? hideOnEmpty;
  final bool? hideOnError;
  final bool? hideSuggestionsOnKeyboardHide;
  final bool? keepSuggestionsOnLoading;
  final bool? keepSuggestionsOnSuggestionSelected;
  final bool? autoFlipDirection;

  PlatformTypeAhead({
    Key? key,
    required this.suggestionsCallback,
    required this.onSuggestionSelected,
    required this.itemBuilder,
    this.suggestionsBoxDecoration,
    this.suggestionsBoxController,
    this.debounceDuration,
    this.loadingBuilder,
    this.noItemsFoundBuilder,
    this.errorBuilder,
    this.transitionBuilder,
    this.animationDuration,
    this.direction,
    this.animationStart = 0.25,
    this.textFieldConfiguration,
    this.suggestionsBoxVerticalOffset = 5.0,
    this.getImmediateSuggestions,
    this.hideOnLoading,
    this.hideOnEmpty,
    this.hideOnError,
    this.hideSuggestionsOnKeyboardHide,
    this.keepSuggestionsOnLoading,
    this.keepSuggestionsOnSuggestionSelected,
    this.autoFlipDirection,
  }) : super(key: key);

  @override
  PlatformTypeAheadState<T> createState() => PlatformTypeAheadState<T>();
}

class PlatformTypeAheadState<T> extends PlatformWidgetState<
    CupertinoTypeAheadField<T>,
    TypeAheadField<T>,
    PlatformTypeAhead<T>> with LoggingMixin {
  late PlatformSuggestionBoxController _controller;

  @override
  TypeAheadField<T> createMaterialWidget(BuildContext context) {
    return TypeAheadField<T>(
        suggestionsCallback: (term) {
          return widget.suggestionsCallback(term);
        },
        itemBuilder: (context, item) {
          return widget.itemBuilder.call(context, item);
        },
        onSuggestionSelected: (item) {
          return widget.onSuggestionSelected.call(item);
        },
        textFieldConfiguration: widget.textFieldConfiguration?.android ??
            const TextFieldConfiguration(),
        suggestionsBoxDecoration:
            widget.suggestionsBoxDecoration ?? const SuggestionsBoxDecoration(),
        debounceDuration:
            widget.debounceDuration ?? const Duration(milliseconds: 300),
        suggestionsBoxController: _controller.android,
        loadingBuilder: widget.loadingBuilder,
        noItemsFoundBuilder: widget.noItemsFoundBuilder,
        errorBuilder: (context, obj) {
          return widget.errorBuilder?.call(context, obj) ?? const SizedBox();
        },
        transitionBuilder: widget.transitionBuilder,
        animationStart: widget.animationStart,
        animationDuration:
            widget.animationDuration ?? const Duration(milliseconds: 500),
        getImmediateSuggestions: widget.getImmediateSuggestions ?? false,
        suggestionsBoxVerticalOffset: widget.suggestionsBoxVerticalOffset,
        direction: widget.direction ?? AxisDirection.down,
        hideOnLoading: widget.hideOnLoading ?? false,
        hideOnEmpty: widget.hideOnEmpty ?? false,
        hideOnError: widget.hideOnError ?? false,
        hideSuggestionsOnKeyboardHide:
            widget.hideSuggestionsOnKeyboardHide ?? true,
        keepSuggestionsOnLoading: widget.keepSuggestionsOnLoading ?? true,
        keepSuggestionsOnSuggestionSelected:
            widget.keepSuggestionsOnSuggestionSelected ?? false,
        autoFlipDirection: widget.autoFlipDirection ?? false);
  }

  @override
  CupertinoTypeAheadField<T> createCupertinoWidget(BuildContext context) {
    return CupertinoTypeAheadField<T>(
        suggestionsCallback: (term) {
          return widget.suggestionsCallback.call(term);
        },
        itemBuilder: (context, item) {
          return widget.itemBuilder.call(context, item);
        },
        onSuggestionSelected: (item) {
          return widget.onSuggestionSelected.call(item);
        },
        textFieldConfiguration: widget.textFieldConfiguration?.ios ??
            const CupertinoTextFieldConfiguration(),
        suggestionsBoxDecoration:
            widget.suggestionsBoxDecoration?.toCupertino() ??
                const CupertinoSuggestionsBoxDecoration(),
        debounceDuration:
            widget.debounceDuration ?? const Duration(milliseconds: 300),
        suggestionsBoxController: _controller.ios,
        loadingBuilder: widget.loadingBuilder,
        noItemsFoundBuilder: widget.noItemsFoundBuilder,
        errorBuilder: widget.errorBuilder == null
            ? null
            : (context, err) => widget.errorBuilder!(context, err),
        transitionBuilder: widget.transitionBuilder,
        animationStart: widget.animationStart,
        animationDuration:
            widget.animationDuration ?? const Duration(milliseconds: 500),
        getImmediateSuggestions: widget.getImmediateSuggestions ?? false,
        suggestionsBoxVerticalOffset: widget.suggestionsBoxVerticalOffset,
        direction: widget.direction ?? AxisDirection.down,
        hideOnLoading: widget.hideOnLoading ?? false,
        hideOnEmpty: widget.hideOnEmpty ?? false,
        hideOnError: widget.hideOnError ?? false,
        hideSuggestionsOnKeyboardHide:
            widget.hideSuggestionsOnKeyboardHide ?? true,
        keepSuggestionsOnLoading: widget.keepSuggestionsOnLoading ?? true,
        keepSuggestionsOnSuggestionSelected:
            widget.keepSuggestionsOnSuggestionSelected ?? false,
        autoFlipDirection: widget.autoFlipDirection ?? false);
  }

  @override
  void initState() {
    super.initState();
    _controller =
        widget.suggestionsBoxController ?? PlatformSuggestionBoxController();
  }
}

class PlatformSuggestionBoxController with LoggingMixin {
  CupertinoSuggestionsBoxController? _ios;

  CupertinoSuggestionsBoxController? get ios {
    _ios ??= CupertinoSuggestionsBoxController();
    return _ios;
  }

  SuggestionsBoxController? _android;

  SuggestionsBoxController? get android {
    _android ??= SuggestionsBoxController();
    return _android;
  }

  PlatformSuggestionBoxController() {
    log.info("Init sbc");
  }
}

extension SuggestionsBoxDecorationExt on SuggestionsBoxDecoration {
  CupertinoSuggestionsBoxDecoration toCupertino() {
    return CupertinoSuggestionsBoxDecoration(
        hasScrollbar: this.hasScrollbar, constraints: this.constraints);
  }
}

class PlatformTextFieldConfiguration<T> {
  /// The decoration to show around the text field.
  ///
  /// Same as [TextField.decoration](https://docs.flutter.io/flutter/material/TextField/decoration.html)
  final InputDecoration? decoration;

  /// The decoration to show around the text field.
  ///
  /// Same as [TextField.decoration](https://docs.flutter.io/flutter/material/TextField/decoration.html)
  final BoxDecoration? boxDecoration;

  /// Controls the text being edited.
  ///
  /// If null, this widget will create its own [TextEditingController](https://docs.flutter.io/flutter/widgets/TextEditingController-class.html).
  /// A typical use case for this field in the TypeAhead widget is to set the
  /// text of the widget when a suggestion is selected. For example:
  ///
  /// ```dart
  /// final _controller = TextEditingController();
  /// ...
  /// ...
  /// TypeAheadField(
  ///   controller: _controller,
  ///   ...
  ///   ...
  ///   onSuggestionSelected: (suggestion) {
  ///     _controller.text = suggestion['city_name'];
  ///   }
  /// )
  /// ```
  final TextEditingController? controller;

  /// Controls whether this widget has keyboard focus.
  ///
  /// Same as [TextField.focusNode](https://docs.flutter.io/flutter/material/TextField/focusNode.html)
  final FocusNode? focusNode;

  /// The style to use for the text being edited.
  ///
  /// Same as [TextField.style](https://docs.flutter.io/flutter/material/TextField/style.html)
  final TextStyle? style;

  /// How the text being edited should be aligned horizontally.
  ///
  /// Same as [TextField.textAlign](https://docs.flutter.io/flutter/material/TextField/textAlign.html)
  final TextAlign textAlign;

  /// Same as [TextField.textDirection](https://docs.flutter.io/flutter/material/TextField/textDirection.html)
  ///
  /// Defaults to null
  final TextDirection? textDirection;

  /// If false the textfield is "disabled": it ignores taps and its
  /// [decoration] is rendered in grey.
  ///
  /// Same as [TextField.enabled](https://docs.flutter.io/flutter/material/TextField/enabled.html)
  final bool? enabled;

  /// The type of keyboard to use for editing the text.
  ///
  /// Same as [TextField.keyboardType](https://docs.flutter.io/flutter/material/TextField/keyboardType.html)
  final TextInputType? keyboardType;

  /// Whether this text field should focus itself if nothing else is already
  /// focused.
  ///
  /// Same as [TextField.autofocus](https://docs.flutter.io/flutter/material/TextField/autofocus.html)
  final bool autofocus;

  /// Optional input validation and formatting overrides.
  ///
  /// Same as [TextField.inputFormatters](https://docs.flutter.io/flutter/material/TextField/inputFormatters.html)
  final List<TextInputFormatter>? inputFormatters;

  /// Whether to enable autocorrection.
  ///
  /// Same as [TextField.autocorrect](https://docs.flutter.io/flutter/material/TextField/autocorrect.html)
  final bool autocorrect;

  /// The maximum number of lines for the text to span, wrapping if necessary.
  ///
  /// Same as [TextField.maxLines](https://docs.flutter.io/flutter/material/TextField/maxLines.html)
  final int maxLines;

  /// The maximum number of characters (Unicode scalar values) to allow in the
  /// text field.
  ///
  /// Same as [TextField.maxLength](https://docs.flutter.io/flutter/material/TextField/maxLength.html)
  final int? maxLength;

  /// If true, prevents the field from allowing more than [maxLength]
  /// characters.
  ///
  /// Same as [TextField.maxLengthEnforced](https://docs.flutter.io/flutter/material/TextField/maxLengthEnforced.html)
  final MaxLengthEnforcement maxLengthEnforcement;

  /// Whether to hide the text being edited (e.g., for passwords).
  ///
  /// Same as [TextField.obscureText](https://docs.flutter.io/flutter/material/TextField/obscureText.html)
  final bool obscureText;

  /// Called when the text being edited changes.
  ///
  /// Same as [TextField.onChanged](https://docs.flutter.io/flutter/material/TextField/onChanged.html)
  final ValueChanged<T>? onChanged;

  /// Called when the user indicates that they are done editing the text in the
  /// field.
  ///
  /// Same as [TextField.onSubmitted](https://docs.flutter.io/flutter/material/TextField/onSubmitted.html)
  final ValueChanged<T>? onSubmitted;

  /// The color to use when painting the cursor.
  ///
  /// Same as [TextField.cursorColor](https://docs.flutter.io/flutter/material/TextField/cursorColor.html)
  final Color? cursorColor;

  /// How rounded the corners of the cursor should be. By default, the cursor has a null Radius
  ///
  /// Same as [TextField.cursorRadius](https://docs.flutter.io/flutter/material/TextField/cursorRadius.html)
  final Radius cursorRadius;

  /// How thick the cursor will be.
  ///
  /// Same as [TextField.cursorWidth](https://docs.flutter.io/flutter/material/TextField/cursorWidth.html)
  final double cursorWidth;

  /// The appearance of the keyboard.
  ///
  /// Same as [TextField.keyboardAppearance](https://docs.flutter.io/flutter/material/TextField/keyboardAppearance.html)
  final Brightness? keyboardAppearance;

  /// Called when the user submits editable content (e.g., user presses the "done" button on the keyboard).
  ///
  /// Same as [TextField.onEditingComplete](https://docs.flutter.io/flutter/material/TextField/onEditingComplete.html)
  final VoidCallback? onEditingComplete;

  /// Configures padding to edges surrounding a Scrollable when the Textfield scrolls into view.
  ///
  /// Same as [TextField.scrollPadding](https://docs.flutter.io/flutter/material/TextField/scrollPadding.html)
  final EdgeInsets scrollPadding;

  /// Configures how the platform keyboard will select an uppercase or lowercase keyboard.
  ///
  /// Same as [TextField.TextCapitalization](https://docs.flutter.io/flutter/material/TextField/textCapitalization.html)
  final TextCapitalization textCapitalization;

  /// The type of action button to use for the keyboard.
  ///
  /// Same as [TextField.textInputAction](https://docs.flutter.io/flutter/material/TextField/textInputAction.html)
  final TextInputAction? textInputAction;

  final EdgeInsetsGeometry padding;
  final String? placeholder;
  final Widget? prefix;
  final OverlayVisibilityMode prefixMode;
  final Widget? suffix;
  final OverlayVisibilityMode suffixMode;
  final OverlayVisibilityMode clearButtonMode;

  const PlatformTextFieldConfiguration({
    this.controller,
    this.focusNode,
    this.decoration,
    this.padding = const EdgeInsets.all(6.0),
    this.placeholder,
    this.prefix,
    this.prefixMode = OverlayVisibilityMode.always,
    this.suffix,
    this.suffixMode = OverlayVisibilityMode.always,
    this.clearButtonMode = OverlayVisibilityMode.never,
    this.keyboardType,
    this.boxDecoration,
    this.textInputAction,
    this.textCapitalization = TextCapitalization.none,
    this.style,
    this.textAlign = TextAlign.start,
    this.autofocus = false,
    this.obscureText = false,
    this.autocorrect = true,
    this.maxLines = 1,
    this.maxLength,
    this.textDirection,
    this.maxLengthEnforcement = MaxLengthEnforcement.enforced,
    this.onChanged,
    this.onEditingComplete,
    this.onSubmitted,
    this.inputFormatters,
    this.enabled,
    this.cursorWidth = 2.0,
    this.cursorRadius = const Radius.circular(2.0),
    this.cursorColor,
    this.keyboardAppearance,
    this.scrollPadding = const EdgeInsets.all(20.0),
  });

  CupertinoTextFieldConfiguration get ios => CupertinoTextFieldConfiguration(
        controller: controller,
        focusNode: focusNode,
        decoration: boxDecoration!,
        padding: padding,
        placeholder: placeholder,
        prefix: prefix,
        prefixMode: prefixMode,
        suffix: suffix,
        suffixMode: suffixMode,
        clearButtonMode: clearButtonMode,
        keyboardType: keyboardType,
        textInputAction: textInputAction,
        textCapitalization: textCapitalization,
        style: style,
        textAlign: textAlign,
        autofocus: autofocus,
        obscureText: obscureText,
        autocorrect: autocorrect,
        maxLines: maxLines,
        maxLength: maxLength,
        maxLengthEnforcement: maxLengthEnforcement,
        onChanged:
            onChanged == null ? null : (String value) => onChanged!(value as T),
        onEditingComplete: onEditingComplete,
        onSubmitted: onSubmitted == null
            ? null
            : (String value) => onSubmitted!(value as T),
        inputFormatters: inputFormatters,
        enabled: enabled!,
        cursorWidth: cursorWidth,
        cursorRadius: cursorRadius,
        cursorColor: cursorColor,
        keyboardAppearance: keyboardAppearance,
        scrollPadding: scrollPadding,
      );

  TextFieldConfiguration get android => TextFieldConfiguration(
      decoration: decoration!,
      style: style,
      controller: controller,
      onChanged: (value) => onChanged?.call(value as T),
      onSubmitted: (value) => onSubmitted?.call(value as T),
      obscureText: obscureText,
      maxLengthEnforcement: maxLengthEnforcement,
      maxLength: maxLength,
      maxLines: maxLines,
      autocorrect: autocorrect,
      inputFormatters: inputFormatters,
      autofocus: autofocus,
      keyboardType: keyboardType!,
      enabled: enabled!,
      textAlign: textAlign,
      focusNode: focusNode,
      cursorColor: cursorColor,
      cursorRadius: cursorRadius,
      textInputAction: textInputAction,
      textCapitalization: textCapitalization,
      cursorWidth: cursorWidth,
      keyboardAppearance: keyboardAppearance,
      onEditingComplete: onEditingComplete,
      textDirection: textDirection,
      scrollPadding: scrollPadding);
}
