import 'dart:io';

import 'package:chewie/chewie.dart';
import 'package:flutter/cupertino.dart';
import 'package:flutter/material.dart';
import 'package:pfile/pfile_api.dart';
import 'package:sunny_dart/sunny_dart.dart';
import 'package:sunny_forms/dividing_line.dart';
import 'package:sunny_forms/icon/image_widgets.dart';
import 'package:video_player/video_player.dart';

/// Video player that wraps [Chewie] and adds some completion handlers.
class StatefulVideoPlayer extends StatefulWidget {
  final String? videoUrl;
  final PFile? videoFile;
  final String? thumbnailUrl;
  final VoidCallback? onCompletion;
  final VideoInitializer? initializer;
  final bool autoplay;
  final bool? showControls;
  final bool isThumbnail;

  const StatefulVideoPlayer({
    Key? key,
    this.videoUrl,
    this.videoFile,
    required this.thumbnailUrl,
    this.onCompletion,
    required this.autoplay,
    this.showControls,
    this.initializer,
    this.isThumbnail = false,
  }) : super(key: key);

  @override
  State<StatefulWidget> createState() => _VideoState();
}

class _VideoState extends State<StatefulVideoPlayer> with LoggingMixin {
  VideoPlayerController? _controller;
  ChewieController? _chewieController;

  bool? _playing;
  int _attempts = 0;
  bool _isReady = false;

  @override
  void initState() {
    super.initState();
    initializeIfNeeded();
  }

  @override
  void didChangeDependencies() {
    initializeIfNeeded();
    super.didChangeDependencies();
  }

  bool _isExpanded = false;

  void initializeIfNeeded() {
    if (widget.videoFile?.path == null && widget.videoUrl == null) {
      return;
    }
    if (_controller == null) {
      _controller = (widget.videoUrl != null
          ? VideoPlayerController.network(widget.videoUrl!)
          : VideoPlayerController.file(File(widget.videoFile!.path!)))
        ..addListener(_onChange);
    }
    var placeholder = widget.thumbnailUrl != null
        ? Image(image: imageProviderOf(widget.thumbnailUrl)!)
        : Container(
            decoration: BoxDecoration(
              color: sunnyColors.g100,
              borderRadius: SunnySpacing.twelvePxRadiusAll,
            ),
          );

    if (_chewieController != null) return;
    this._chewieController = ChewieController(
      videoPlayerController: _controller!,
      autoPlay: widget.isThumbnail ? false : widget.autoplay,
      autoInitialize: false,
      allowFullScreen: false,
      allowPlaybackSpeedChanging: false,
      // customControls: ,
      // overlay: widget.isThumbnail ? emptyBox : null,
      // customControls: widget.isThumbnail ? emptyBox : null,
      // aspectRatio: 16.0 / 9.0,
      // ignore: avoid_bool_literals_in_conditional_expressions
      showControls: _isExpanded == true ||
          (widget.isThumbnail
              ? false
              : (widget.autoplay != true)
                  ? true
                  : widget.showControls ?? true),
      showControlsOnInitialize: false,
      placeholder: placeholder,
      errorBuilder: (err, stack) {
        log.severe("Error with video $err", err);
        return Container(
            child: ErrorText("There was a problem loading your video"));
      },
      allowMuting: true,
      // materialProgressColors: ChewieProgressColors(
      //     playedColor: Colors.amber,
      //     handleColor: Colors.orange,
      //     backgroundColor: Colors.blueGrey,
      //     bufferedColor: Colors.brown),
      // customControls: MaterialControls(),
    );

    _controller!.initialize().then((_) {
      _playing = true;
      widget.initializer?.call(_chewieController);
      if (mounted) {
        setState(() {
          this._isReady = true;
        });
      }
    }).catchError((e, st) {
      log.info(e, st);
    });
  }

  @override
  void dispose() {
    _controller?.dispose();
    _chewieController?.dispose();
    super.dispose();
  }

  _onChange() async {
    if (_controller == null) return;
    final videoState = _controller!.value;
    if (videoState.hasError) {
      if (_attempts++ <= 3) {
        // Reset
        log.info("VIDEO: HAD ERROR: ${videoState.errorDescription}");
      } else {
        log.info("VIDEO: FATAL: ${videoState.errorDescription}");
      }
    }
    if (_playing == true &&
        videoState.position.inMilliseconds >=
            (videoState.duration.inMilliseconds)) {
      setState(() {
        log.info("VIDEO: DONE");
        _playing = false;
        _controller!.pause();
        widget.onCompletion?.call();
      });
    }
  }

  @override
  Widget build(BuildContext context) {
    if (_chewieController != null && _isReady) {
      return Chewie(controller: _chewieController!);
    } else {
      return emptyBox;
    }
  }

// playedColor: Theme.of(context).accentColor,
//                  handleColor: Theme.of(context).accentColor,
//                  bufferedColor: Theme.of(context).backgroundColor,
//                  backgroundColor: Theme.of(context).disabledColor)
//  final theme = ThemeData();
//  return Material(
//  textStyle: TextStyle(color: Colors.amber),
//  child: Theme(
//  child: Container(
//  margin: EdgeInsets.only(top: 5, bottom: 5),
//  decoration: BoxDecoration(borderRadius: BorderRadius.all(Radius.circular(20.0))),
//  child: Chewie(controller: _chewieController)),
//  data: theme.copyWith(
//  buttonColor: Colors.black,
//  textTheme: theme.textTheme.apply(
//  bodyColor: Colors.purpleAccent,
//  displayColor: Colors.purpleAccent,
//  decorationColor: Colors.brown,
//  ),
//  dialogBackgroundColor: Colors.purple,
//  ),
//  ),
//  );
}

typedef VideoInitializer = void Function(ChewieController? controller);
