import 'package:dartxx/dartxx.dart';
import 'package:flutter/cupertino.dart';
import 'package:sunny_platform_widgets/sunny_platform_widgets.dart';
import 'package:sunny_forms/attribute_form_field.dart';
import 'package:sunny_forms/sunny_form_controller.dart';
import 'package:sunny_forms/sunny_form_field.dart';
import 'package:sunny_forms/text_widgets.dart';

typedef OnSubmitHandler<T> = Future<String> Function(T data);

Future<String?> showSimpleTextEntryDialog(
  BuildContext context, {
  Key? key,
  required String title,
  String? subtitle,
  String? saveButtonLabel,
  SunnyFormFieldBuilder<dynamic>? textFieldBuilder,
  FormFieldValidator? validator,
  String? textFieldValue,
  OnSubmitHandler<String>? onSubmit,
}) {
  return showPlatformDialog(
      context: context,
      builder: (context) => SimpleTextEntryDialog(
            key: key,
            title: title,
            subtitle: subtitle,
            saveButtonLabel: saveButtonLabel,
            textFieldBuilder: textFieldBuilder,
            textFieldValue: textFieldValue,
            validator: validator,
            onSubmit: onSubmit,
          ));
}

class SimpleTextEntryDialog extends StatefulWidget {
  final String title;
  final String? subtitle;
  final String? saveButtonLabel;
  final SunnyFormFieldBuilder<dynamic>? textFieldBuilder;
  final FormFieldValidator<String>? validator;
  final String? textFieldValue;
  final OnSubmitHandler<String>? onSubmit;

  const SimpleTextEntryDialog({
    required this.title,
    Key? key,
    this.subtitle,
    this.saveButtonLabel,
    this.textFieldBuilder,
    this.validator,
    this.textFieldValue,
    required this.onSubmit,
  }) : super(key: key);

  @override
  State<StatefulWidget> createState() => SimpleTextEntryState();
}

final textValuePath = JsonPath.of("textValue");

class SimpleTextEntryState extends State<SimpleTextEntryDialog> {
  bool _submitting = false;
  String? onSubmitError;
  TextEditingController? textController;
  late Map<String, dynamic> values;

  @override
  void initState() {
    super.initState();
    values = <String, dynamic>{"textValue": widget.textFieldValue};
    textController = TextEditingController(text: widget.textFieldValue);
  }

  @override
  void dispose() {
    textController!.dispose();
    super.dispose();
  }

  SunnyFormFieldBuilder defaultBuilder() {
    return (BuildContext context, SunnyFormFieldState state) {
      return PlatformTextField(
        autofocus: true,
        controller: textController,
        onChanged: (str) {
          state.updateValue(str, AttributeSource.control);
        },
        onSubmitted: (str) {
          state.updateValue(str, AttributeSource.control);
        },
        textCapitalization: TextCapitalization.none,
        expands: false,
      );
    };
  }

  @override
  Widget build(BuildContext context) {
    final builder = widget.textFieldBuilder ?? defaultBuilder();
    final intl = SunnyIntl.of(context);

    return SunnyForm.builder(
      "simpleTextEntryModal",
      form: (context) => FormController.ofMap(context, values,
          validators: {
            textValuePath: (value) {
              return widget.validator?.call(value?.toString());
            }
          },
          name: "simpleTextEntryModal"),
      builder: (context, controller) => PlatformAlertDialog(
        title: Text(widget.title),
        content: Layout.column().min.build([
          if (widget.subtitle != null) Body2Text(widget.subtitle!),
          verticalSpace,
          if (onSubmitError != null) ...[
            ErrorText(
              onSubmitError!,
              isCentered: true,
            ),
          ],
          AttributeFormField(
            attribute: "textValue",
            isRequired: true,
            builder: (c, st) {
              return builder(c, st);
            },
          ),
        ]),
        actions: [
          PlatformDialogAction(
            key: Key("text-entry-cancel"),
            child: Text(intl!.cancelLabel),
            onPressed: _submitting
                ? null
                : () {
                    Navigator.pop(context, null);
                  },
          ),
          PlatformDialogAction(
            key: Key("text-entry-save"),
            child: Text(widget.saveButtonLabel ?? intl.saveLabel),
            cupertino: (context, _) =>
                CupertinoDialogActionData(isDefaultAction: true),
            onPressed: _submitting
                ? null
                : () async {
                    setState(() {
                      _submitting = true;
                    });
                    try {
                      final result = await controller.submit<String>(context,
                          submit: () async {
                        final errorMsg = (await widget.onSubmit
                            ?.call(this.textController!.text));
                        if (errorMsg?.isEmpty != false) {
                          return null;
                        } else {
                          return errorMsg;
                        }
                      });

                      if (result.isSuccessful &&
                          result.value?.isNotEmpty != true) {
                        Navigator.pop(context, controller.get(textValuePath));
                      } else {
                        setState(() {
                          this.onSubmitError = result.value;
                        });
                      }
                    } finally {
                      setState(() {
                        _submitting = false;
                      });
                    }
                  },
          )
        ],
      ),
    );
  }
}
