import 'dart:math' as math;

import 'package:flutter/material.dart';
import 'package:overlay_support/overlay_support.dart';
import 'package:percent_indicator/percent_indicator.dart';
import 'package:sunny_dart/helpers/logging_mixin.dart';
import 'package:sunny_forms/standard_column.dart';
import 'package:sunny_sdk_core/model/progress_tracker.dart';

typedef ProgressBuilder = Widget Function(double progress);

class SunnyProgressIndicator<T> extends StatefulWidget {
  final ProgressTracker<T>? tracker;
  final double? size;
  final bool circular;
  final bool showPercent;
  final bool showTask;
  final Color? background;
  final ProgressBuilder? builder;

  const SunnyProgressIndicator({
    Key? key,
    required this.tracker,
    this.size,
    this.circular = true,
    this.showPercent = false,
    this.showTask = false,
    this.background,
    this.builder,
  }) : super(key: key);

  @override
  _SunnyProgressIndicatorState<T> createState() =>
      _SunnyProgressIndicatorState<T>();
}

class _SunnyProgressIndicatorState<T> extends State<SunnyProgressIndicator<T>>
    with LoggingMixin {
  bool get _showTask {
    return widget.showTask && widget.tracker!.task!.isNotEmpty == true;
  }

  double? _pct = 0.0;

  ProgressTracker<T>? _tracker;

  @override
  Widget build(BuildContext context) {
    log.info("Building Tracker! ${_pct}");
    final disabledColor = Theme.of(context).unselectedWidgetColor;
    final _size = widget.size ?? 50.0;
    if (widget.circular) {
      final stroke = math.min(10.0, _size / 10.0);

      final circ = CircularPercentIndicator(
        key: _tracker!.key is Key ? _tracker!.key as Key? : ValueKey(_tracker!.key),
        radius: _size,
        lineWidth: stroke,
        animation: true,
        animationDuration: 200,
        animateFromLastPercent: true,
        backgroundColor: widget.background ?? disabledColor,
        fillColor: widget.background ?? disabledColor,
        circularStrokeCap: CircularStrokeCap.round,
        center: widget.builder?.call(_tracker!.progress) ??
            CircleAvatar(
              // Separates the stroke from the center by equal amount
              radius: _size - (4.0 * stroke),
              backgroundColor: sunnyColors.primaryColor,
            ),
        percent: _pct!,
        progressColor: Theme.of(context).primaryColor,
      );
      if (_showTask || widget.showPercent) {
        return centerColumn([
          if (widget.showTask && _tracker!.task!.isNotEmpty == true)
            Text(_tracker!.task!),
          circ,
          if (widget.showPercent == true) Text(_tracker!.percentText),
        ], color: Colors.transparent);
      } else {
        return circ;
      }
    } else {
      return centerColumn([
        if (widget.showTask && _tracker!.task!.isNotEmpty == true)
          Text(_tracker!.task!),
        LinearPercentIndicator(
          key: _tracker!.key is Key
              ? _tracker!.key as Key?
              : ValueKey(_tracker!.key),
          width: _size,
          lineHeight: 10,
          animation: true,
          animationDuration: 200,
          animateFromLastPercent: true,
          backgroundColor: widget.background ?? disabledColor,
          linearStrokeCap: LinearStrokeCap.roundAll,
          percent: _tracker!.percent,
          progressColor: Theme.of(context).primaryColor,
        ),
        if (widget.showPercent) Text(_tracker!.percentText),
      ]);
    }
  }

  @override
  void initState() {
    super.initState();
    _tracker = widget.tracker;
    _pct = _tracker?.percent ?? 0.0;
    widget.tracker!.addListener(_onChange);
    log.info("Added tracker");
  }

  @override
  void didUpdateWidget(SunnyProgressIndicator<T> oldWidget) {
    if (widget.tracker != _tracker) {
      log.info("Got update from tracker");
      if (_tracker?.isDisposed != true) {
        _tracker?.removeListener(_onChange);
      }
      _tracker = widget.tracker;
      _tracker?.addListener(_onChange);
    }
    super.didUpdateWidget(oldWidget);
  }

  @override
  void dispose() {
    log.info("Disposed tracker");
    if (_tracker?.isDisposed != true) {
      _tracker!.removeListener(_onChange);
    }
    super.dispose();
  }

  void _onChange() {
    setState(() {
      this._pct = _tracker?.percent;
    });
  }
}

Future showProgressOverlay(ProgressTracker tracker, Future completion()) async {
  OverlaySupportEntry? overlayEntry;
  overlayEntry = showOverlay((context, progress) {
    return GestureDetector(
      onTap: () {
        overlayEntry?.dismiss(animate: true);
      },
      child: SunnyProgressIndicator(tracker: tracker),
    );
  }, duration: Duration.zero);

  try {
    await completion();
  } finally {
    overlayEntry.dismiss(animate: true);
  }
}
