import 'dart:async';

import 'package:flutter/widgets.dart';
import 'package:meta_forms/mschema_service.dart';
import 'package:mobx/mobx.dart' hide ObservableMap;
import 'package:observable_collections/observable_collections.dart';
import 'package:provider/provider.dart';
import 'package:sunny_dart/sunny_dart.dart' hide Consumer;
import 'package:sunny_forms/builders/event_bridge.dart';
import 'package:sunny_forms/sunny_hud.dart';
import 'package:sunny_sdk_core/sunny_sdk_core.dart';
import 'package:sunny_service_stubs/meta.dart';
import 'package:sunny_service_stubs/models.dart';

typedef FormValidator = FutureOr<List<ValidationError>> Function(
    FormController form);
typedef FormSubmission<T> = FutureOr<T> Function(FormController form);
typedef PropertyGetter = dynamic Function(JsonPath path);
typedef PropertySetter = void Function(JsonPath path, dynamic);
typedef AttributeChanged = void Function(JsonPath path, dynamic value);
typedef SubmitHook = FutureOr<Result> Function(
    FormController controller, FormSubmitResultBuilder resultBuilder);
enum SunnyFormState { ready, validating, submitting, error }

typedef ResultExtractor<T> = T Function(FormSubmitResult result);
typedef FormWidgetBuilder<T> = Widget Function(
  BuildContext context,
  FormController<T> form,
);

/// A data form base class.  Similar to [Form], but adds more support for things like nested forms, validation, etc.
///
/// Use one of the static constructors:
/// [FormController.singleValue] when the form is a single value
/// [FormController.ofModel] when the form is backed by a single [MModel] instance
/// [FormController.ofMSchema] when the form is backed by a [MSchemaDefinition]
/// [FormController.ofMSchemaFuture] when the [MSchemaDefinition] is not loaded yet
/// [FormController.ofMap] key/value pairs
///
abstract class FormController<T>
    with Disposable, LoggingMixin
    implements IFormController {
  /// Whether the user has made changes to the form.
  bool get isModified;

  /// An observable map of errors
  SunnyObservableMap<JsonPath, ValidationError?> get errors;

  /// The parent form
  FormController? get parent;

  /// The build context for the form
  BuildContext get buildContext;

  /// Private setter
  set _parent(FormController? parent);

  /// Helpful for debugging - displays the name of the record being edited, for example
  String get name;

  Future dispose();

  void autofocus([BuildContext? context]);

  void focusLast([BuildContext? context]);

  /// Validators run during the validation phase of the submission.
  Map<JsonPath, FormFieldValidator> get validators;

  /// Consumers can request a completer that's used to prevent submission until the child component is
  /// ready - maybe they are uploading a file or something.
  Map<JsonPath, SafeCompleter> get readyChecks;

  /// A stream of any change to the state of this form.
  SyncStream<SunnyFormState> get formState;

  /// Retrieves a value from the underlying data model.
  X? get<X>(JsonPath path);

  /// Retrieves or sets the value from the underlying data model, but uses the type parameter from the [JsonPath]
  X? path<X>(JsonPath<X> path, [Attribute<X>? value]);

  /// Sets a value in the underlying model
  void set(JsonPath path, value,
      {AttributeSource source = AttributeSource.model});

  /// Put is like [set], but it's not assumed that the data has been sanitized
  void put(JsonPath path, value, {required AttributeSource source});

  /// Registers a subform
  void registerForm(FormController? form);

  /// Unregisters a subform
  void unregisterForm(FormController form);

  /// Gets or puts a ready check at the given path
  SafeCompleter requestReadyCheck(JsonPath path);

  /// Adds a focusNode to this form at this path
  void addFocusNode(JsonPath path, FocusNode? node);

  /// Retreives a focus node for the specified path
  FocusNode? getFocusNode(JsonPath path, {bool createIfMissing = true});

  /// Returns the focusNode that should be autofocused based on the current form state.  It's generally the first
  /// required form field that doesn't have a value
  FocusNode? get autofocusNode;

  /// The last focus node that was added
  FocusNode? get lastFocusNode;

  /// Unregisters a ready check at the given path.
  /// (should this also complete any outstanding item?)
  void invalidateReadyCheck(JsonPath path);

  /// Submits this form, including subforms, and returns the result
  FutureOr<FormSubmitResult<X>> submit<X>(BuildContext context,
      {FutureOr<X?> submit()?, bool? skipIfUnmodified});

  // TODO(ericm): Review this
  String? validatePath(JsonPath path);

  /// Validates this form and any subforms.  An empty result means that the form validated successfully
  FutureOr<Map<JsonPath, List<ValidationError>>> validate(
      {required bool skipIfUnmodified});

  /// Begins this forms submission process
  void startSubmission() {
    formState.update(SunnyFormState.submitting);
  }

  /// Ends this form's submission process
  void endSubmission() {
    formState.update(SunnyFormState.ready);
  }

  /// Registers an error at this path
  void addError(JsonPath path, ValidationError? error,
      {AttributeSource source = AttributeSource.model});

  /// Removes all errors at this [path]
  void clearError(JsonPath path, {required AttributeSource source});

  /// Clears out all errors
  void clearErrors({required AttributeSource source});

  /// Clears and sets a list of [validationErrors]
  void updateErrors(List<ValidationError> validationErrors,
      {required AttributeSource source});

  /// Observes any new values (including null) for a single [path]
  Stream<X?> observePath<X>(JsonPath path) => stream.where((attr) {
        return attr.path == path;
      }).map((attr) => attr.value as X?);

  /// observes changes to all paths
  Stream<Attribute> get stream;

  /// Registers a submit hook.  These run after validation
  void registerSubmitHook(String key, SubmitHook? onSubmit);

  /// Unregisters a submit hook
  void unregisterSubmitHook(String key);

  /// Unregisters a validator
  void unregisterFormValidator(String key);

  /// Registers a form validator
  void registerFormValidator(String key, FormValidator? validator);

  static SingleValueForm<T> singleValue<T>(
    BuildContext context,
    SingleValue<T> container, {
    Map<JsonPath, FormFieldValidator>? validators,
    required name,
  }) {
    return SingleValueForm<T>(context, container, validators, name: name);
  }

  static MModelForm<T> ofModel<T extends MBaseModel>(
    BuildContext context,
    T model, {
    Map<JsonPath, FormFieldValidator>? validators,
  }) {
    return MModelForm(context, model,
        mtype: model.mtype, validators: validators);
  }

  static FormController ofMSchema<T extends MBaseModel>(
    BuildContext context,
    T model,
    IMSchemaDefinition definition, {
    Map<JsonPath, FormFieldValidator>? validators,
  }) {
    return MSchemaForm<T>(context, model, definition.asEntitySchema(),
        validators: validators);
  }

  static Future<MSchemaForm<T>> ofMSchemaFuture<T extends MModel>(
    BuildContext context,
    T model, {
    MSchemaRef? mtype,
    Map<JsonPath, FormFieldValidator>? validators,
  }) async {
    assert(mtype != null);
    final definition =
        await context.mschemaService.getSchema(mtype ?? model.mtype);
    return MSchemaForm<T>(context, model, definition.asEntitySchema(),
        validators: validators);
  }

  static MapForm ofMap(BuildContext context, Map<String, dynamic> map,
      {Map<JsonPath, FormFieldValidator>? validators, required name}) {
    return MapForm(context, map, validators, name: name);
  }
}

abstract class SunnyFormController<T>
    with LoggingMixin, Store, Disposable
    implements FormController<T> {
  SunnyFormController.base(
    this.buildContext,
    Map<JsonPath, FormFieldValidator>? validators,
    name,
  )   : name = name?.toString() ?? "[unknown]",
        validators = validators ?? {};

  /// Whether the user has made changes to the form.
  bool _isModified = false;

  @override
  final BuildContext buildContext;

  @override
  bool get isModified {
    return _isModified || _subforms.any((form) => form!.isModified);
  }

  /// An observable map of errors
  @override
  final errors = SunnyObservableMap<JsonPath, ValidationError>();

  /// Subforms are nested independent forms that wish to be tied to the parent form lifecycle
  final List<FormController?> _subforms = [];

  /// If this form is a child form, [parent] stores the reference to the parent form
  FormController? _parent;

  @override
  FormController? get parent => _parent;

  /// Helpful for debugging - displays the name of the record being edited, for example
  @override
  final String name;

  final StreamController<Attribute> _changeStream =
      StreamController.broadcast();

  @override
  Future dispose() async {
    // [Disposable] any participant can register disposal for this form
    await super.disposeAll();
    _parent?.unregisterForm(this);
    await Future.wait([_changeStream.close(), errors.dispose()]);
    _focusNodes.values.forEach((focusNode) {
      // don't dispose??
//      focusNode.dispose();
    });
  }

  dynamic _get(JsonPath? path);

  void _set(JsonPath path, value);

  void _put(JsonPath path, value) => _set(path, value);

  /// Validators run during the validation phase of the submission.
  @override
  final Map<JsonPath, FormFieldValidator> validators;

  final Map<JsonPath, FocusNode?> _focusNodes = {};

  /// Submit hooks are executed after validators run
  final Map<String, SubmitHook> _submitHooks = {};

  /// Submit hooks are executed after validators run
  final Map<String, FormValidator> _formValidators = {};

  /// Consumers can request a completer that's used to prevent submission until the child component is
  /// ready - maybe they are uploading a file or something.
  @override
  final readyChecks = <JsonPath, SafeCompleter>{};

  @override
  final SyncStream<SunnyFormState> formState = SyncStream.controller(
      initialValue: SunnyFormState.ready, debugName: "formState");

  @override
  X? get<X>(JsonPath path) => _get(path) as X?;

  /// Uses the type from the JsonPath
  @override
  X? path<X>(JsonPath<X> path, [Attribute<X>? attr]) {
    if (attr == null) {
      return _get(path) as X?;
    } else {
      _set(path, attr.value);
      _changeStream.add(attr);
      return attr.value;
    }
  }

  @override
  void registerForm(FormController? form) {
    if (_subforms.contains(form)) {
      log.warning("Already registered subform with name $name");
    } else if (form!.parent != null) {
      log.warning(
          "Child form ${form.name} already has a parent ${form.parent!.name}");
    }
    form!._parent = this;
    _subforms.add(form);
  }

  @override
  void unregisterForm(FormController form) {
    form._parent = null;
    _subforms.remove(form);
  }

  @override
  void addFocusNode(JsonPath path, FocusNode? newFocusNode) {
    final previous = _focusNodes[path];
    if (previous != null && previous != newFocusNode) {
      log.warning(
        "Attempting to change an existing focus node to a different value for path $path",
        null,
        StackTrace.current,
      );
      _focusNodes[path]!.dispose();
    }
    _focusNodes[path] = newFocusNode;
  }

  @override
  void autofocus([BuildContext? context]) {
    context ??= this.buildContext;
    final autoFocusNode = this.autofocusNode;
    if (autoFocusNode != null) {
      log.info("Form $name autofocus node ${autoFocusNode.debugLabel}");
//      autoFocusNode.requestFocus();
    } else {
      log.info("Form $name has no autofocus node");
    }
  }

  @override
  void focusLast([BuildContext? context]) {
    context ??= this.buildContext;
    final autoFocusNode = this.lastFocusNode;
    if (autoFocusNode != null) {
      log.info("Form $name last focus node ${autoFocusNode.debugLabel}");
//      final scope = FocusScope.of(context);
//      scope.requestFocus(autoFocusNode);
    } else {
      log.info("Form $name has no last node");
    }
  }

  @override
  FocusNode? get lastFocusNode {
    return _focusNodes.entries.toList().lastOrNull()?.value ??
        _subforms.map((f) => f!.lastFocusNode).firstOr();
  }

  @override
  FocusNode? getFocusNode(JsonPath path, {bool createIfMissing = true}) {
    var existing = _focusNodes[path];
    if (existing == null && createIfMissing) {
      existing = FocusNode(debugLabel: "$name:$path");
      _focusNodes[path] = existing;
    }
    return existing;
  }

  @override
  FocusNode? get autofocusNode {
    final myAutofocusNode = _focusNodes.entries
        .where((entry) {
          final currentValue = get<Object>(entry.key);
          return currentValue.isNullOrBlank;
        })
        .firstOrNull
        ?.value;

    if (myAutofocusNode != null) return myAutofocusNode;
    for (final subform in _subforms) {
      final subformFocus = subform!.autofocusNode;
      if (subformFocus != null) return subformFocus;
    }
    return null;
  }

  @override
  void set(JsonPath path, value,
      {AttributeSource source = AttributeSource.model}) {
    try {
      if (source == AttributeSource.control) {
        _isModified = true;
      }
      _set(path, value);
      _changeStream.add(Attribute(path, value, source: source));

      final validator = validators[path];
      final validationResult = validator?.call(value);
      if (validationResult != null) {
        addError(
            path,
            ValidationError.ofString(
              path,
              validationResult,
              debugMessage: "Custom validator",
            ),
            source: source);
      } else {
        /// Changed this to [model] because typing updates in the form control
        /// wasn't triggering errors to update
        clearError(path, source: source);
      }
    } catch (e) {
      log.warning("Invalid value in form for $path: $e");

      /// In the case of an error
      addError(
        path,
        ValidationError.ofString(path, "Invalid value.", debugMessage: "$e"),
        source: source,
      );
    }
  }

  @override
  SafeCompleter requestReadyCheck(JsonPath path) {
    return readyChecks.putIfAbsent(path, () {
      final completer = SafeCompleter();
      return completer;
    });
  }

  @override
  void invalidateReadyCheck(JsonPath path) {
    final check = readyChecks[path];
    if (check?.isStarted == true && check?.isNotComplete == true) {
      check!.complete();
    }
    readyChecks.remove(path);
  }

  /// Put is like set, but it's not assumed that the data has been sanitized
  @override
  void put(JsonPath path, value, {required AttributeSource source}) {
    _put(path, value);
    _changeStream.add(Attribute(path, value, source: source));
    clearError(path, source: source);
    if (source == AttributeSource.control) {
      _isModified = true;
    }
  }

  @override
  FutureOr<FormSubmitResult<X>> submit<X>(BuildContext context,
      {FutureOr<X?> submit()?, bool? skipIfUnmodified}) async {
    skipIfUnmodified ??= false;

    final start = DateTime.now();
    startSubmission();
    var result = FormSubmitResultBuilder<X>.start(this);

    for (final entry in readyChecks.entries) {
      log.info("Awaiting ready check: ${entry.key}");
      if (entry.value.isActive) {
        await entry.value.future;
        log.info("Finished ready check: ${entry.key}");
      } else {
        log.info("Found ready check, but it was not active: ${entry.key}");
      }
    }

    final hasChanges = skipIfUnmodified != true || isModified;

    /// first, validate all the way down the chain:
    final validationErrors = await validate(skipIfUnmodified: skipIfUnmodified);
    if (validationErrors.isNotEmpty && hasChanges) {
      result
        ..validationErrors = validationErrors
        ..resultType = ResultType.error
        ..message = "Validation error";
      return result.end();
    }

    final copy = [..._subforms];
    for (final subform in copy) {
      try {
        log.info(
            "Submitting subform ${subform!.name}: modified=${subform.isModified} ${skipIfUnmodified != true ? '(saving anyway)' : ''}");

        ///
        /// Submitting each subform
        ///
        final subformResult =
            await subform.submit(context, skipIfUnmodified: skipIfUnmodified);
        log.info("Done with ${subform.name} subform ${subform.name}");
        result.subforms[subform.name] = subformResult;
        if (subformResult.isError) {
          log.warning(
              "Early termination from subform ${subform.name}: ${subformResult.message}");
          result
            ..resultType = ResultType.error
            ..message =
                "Subform error ${subform.name}: ${subformResult.message}";

          return result.end();
        }
      } catch (e) {
        result.subforms[subform!.name] = FormSubmitResult.error(this, e);
        result
          ..resultType = ResultType.error
          ..message = "$e";
        return result.end();
      }
    }

    try {
      // Run onSubmits
      final hooks = [..._submitHooks.entries];
      if (skipIfUnmodified != true || isModified) {
        for (final hook in hooks) {
          log.info("Executing hook: ${hook.key}");
          final hookResult = await hook.value(this, result);
          result.submitHooks[hook.key] = hookResult;
          if (!hookResult.isSuccessful) {
            result
              ..resultType = ResultType.error
              ..message =
                  "Failed submit hook ${hook.key}: ${hookResult.message}";
            return result.end();
          }
          log.info("Hook: ${hook.key}: $hookResult");
        }
        X? submitResult = (await submit?.call());
        result
          ..value = submitResult
          ..resultType = ResultType.success;
      } else {
        result..resultType = ResultType.skipped;
      }
      return result.end();
    } catch (e, stack) {
      if (e is BadRequestException) {
        result.validationErrors = e.validationErrors.groupBy((e) => e.path);
        this.updateErrors(e.validationErrors, source: AttributeSource.model);
        result.message = "Validation Errors";
      } else {
        result.message = "Error submitting form";
        log.info(e);
        log.info(stack);
        SunnyHud.error(context, "Error submitting form");
      }
      result.resultType = ResultType.error;
      return result.end();
    } finally {
      endSubmission();
      log.info(
          "Form submitted in ${DateTime.now().difference(start).inMilliseconds}");
    }
  }

  @override
  String? validatePath(JsonPath path) => errors[path]?.message;

  @override
  FutureOr<Map<JsonPath, List<ValidationError>>> validate(
      {required bool skipIfUnmodified}) async {
    try {
      final subforms = [...this._subforms];

      List<ValidationError> validationErrors = [];
      for (final subform in subforms) {
        try {
          final errs =
              await subform!.validate(skipIfUnmodified: skipIfUnmodified);
          validationErrors += errs.values.flatten();
        } on BadRequestException catch (e) {
          validationErrors += e.validationErrors;
        }
      }

      List<ValidationError> myErrors;
      if (_isModified || skipIfUnmodified != true) {
        // Do the inline validation
        final inlineValidators = validators
            .mapEntries((JsonPath path, FormFieldValidator validator) {
          final response = validator(get(path));
          if (response != null) {
            return ValidationError(path: path, message: response);
          } else {
            return null;
          }
        }).notNull();

        final processingErrors =
            await _formValidators.entries.map((entry) async {
          try {
            final validator = entry.value;
            final errors = await validator.call(this);
            return errors;
          } on BadRequestException catch (e) {
            return e.validationErrors;
          } catch (e, stack) {
            log.severe("An unexpected error while validating: $e", stack);
            return [
              ValidationError.ofString(
                JsonPath.root(),
                "Unknown error validating form",
                debugMessage: "$e",
              ),
            ];
          }
        }).awaitAll();

        myErrors = [
          ...inlineValidators,
          ...processingErrors.flatten(),
        ];

        myErrors.forEach((err) {
          addError(err.path, err, source: AttributeSource.model);
        });
      } else {
        myErrors = [];
      }

      return [...validationErrors, ...myErrors].groupBy(((e) => e.path));
    } catch (e) {
      log.severe("An unexpected error while validating: $e");
      rethrow;
    }
  }

  @override
  void startSubmission() {
    formState.update(SunnyFormState.submitting);
  }

  @override
  void endSubmission() {
    formState.update(SunnyFormState.ready);
  }

  @override
  void addError(JsonPath path, ValidationError? error,
      {AttributeSource source = AttributeSource.model}) {
    errors[path] = error;
    _changeStream.add(Attribute.ofError(error!.path, error, source: source));
  }

  @override
  void clearError(JsonPath path, {required AttributeSource source}) {
    errors.remove(path);
    // pushes a change with the error cleared
    _changeStream
        .add(Attribute(path, get(path), source: source, clearErrors: true));
  }

  @override
  void clearErrors({required AttributeSource source}) {
    errors.keys.notNull().forEach((path) {
      _changeStream
          .add(Attribute(path, get(path), source: source, clearErrors: true));
    });

    errors.clear();
  }

  @override
  void updateErrors(List<ValidationError> validationErrors,
      {required AttributeSource source}) {
    log.info("Got validation errors:");
    validationErrors.forEach((error) {
      log.info("  - ${error.path} ${error.code}: ${error.message}");
    });
    clearErrors(source: source);
    validationErrors.forEach((error) {
      errors[error.path] = error;
      _changeStream.add(Attribute.ofError(error.path, error, source: source));
    });
  }

  /// Observes a single path
  @override
  Stream<X?> observePath<X>(JsonPath path) => stream.where((attr) {
        return attr.path == path;
      }).map((attr) => attr.value as X?);

  /// observes all changes
  @override
  Stream<Attribute> get stream => _changeStream.stream;

  @override
  void registerSubmitHook(String key, SubmitHook? onSubmit) {
    if (onSubmit != null) {
      _submitHooks[key] = onSubmit;
    }
  }

  @override
  void unregisterFormValidator(String key) {
    _formValidators.remove(key);
  }

  @override
  void registerFormValidator(String key, FormValidator? validator) {
    if (validator != null) {
      _formValidators[key] = validator;
    }
  }

  @override
  void unregisterSubmitHook(String key) {
    _submitHooks.remove(key);
  }

  @override
  bool operator ==(Object other) =>
      identical(this, other) ||
      other is FormController &&
          runtimeType == other.runtimeType &&
          name == other.name;

  @override
  int get hashCode => name.hashCode;

  static SingleValueForm<T> singleValue<T>(
    BuildContext context,
    SingleValue<T> container, {
    Map<JsonPath, FormFieldValidator>? validators,
    required name,
  }) {
    return SingleValueForm<T>(context, container, validators, name: name);
  }

  static MModelForm<T> ofModel<T extends MModel>(
    BuildContext context,
    T model, {
    Map<JsonPath, FormFieldValidator>? validators,
  }) {
    return MModelForm(context, model,
        mtype: model.mtype, validators: validators);
  }

  static FormController ofMSchema<T extends MModel>(
    BuildContext context,
    T model,
    IMSchemaDefinition definition, {
    Map<JsonPath, FormFieldValidator>? validators,
  }) {
    return MSchemaForm<T>(context, model, definition.asEntitySchema(),
        validators: validators);
  }

  static Future<MSchemaForm<T>> ofMSchemaFuture<T extends MModel>(
    BuildContext context,
    T model, {
    MSchemaRef? mtype,
    Map<JsonPath, FormFieldValidator>? validators,
  }) async {
    assert(mtype != null);
    final definition =
        await context.mschemaService.getSchema(mtype ?? model.mtype);
    return MSchemaForm<T>(context, model, definition.asEntitySchema(),
        validators: validators);
  }

  static MapForm ofMap(BuildContext context, Map<String, dynamic> map,
      {Map<JsonPath, FormFieldValidator>? validators, required name}) {
    return MapForm(context, map, validators, name: name);
  }
}

class SingleValueForm<V> extends SunnyFormController<V> {
  final SingleValue<V> container;

  SingleValueForm(BuildContext context, this.container,
      Map<JsonPath, FormFieldValidator>? validators,
      {required name})
      : super.base(context, validators, name ?? "singleValue");

  @override
  void _set(JsonPath path, value) => container.update(value as V?);

  @override
  dynamic _get(JsonPath? path) => container.get();
}

class MapForm extends SunnyFormController<Map<String, dynamic>> {
  final Map<String, dynamic> container;

  MapForm(BuildContext context, this.container,
      Map<JsonPath, FormFieldValidator>? validators,
      {required name})
      : super.base(context, validators, name ?? "mapForm");

  @override
  void _set(JsonPath path, value) => Maps.setByPath(container, path, value);

  @override
  dynamic _get(JsonPath? path) => Maps.getByPath(container, path!);

  @override
  String get name => "mapForm";
}

abstract class ModelBasedFormController<T extends MBaseModel>
    implements SunnyFormController<T> {
  T get mmodel;
}

class MBaseModelFormController<T extends MBaseModel>
    extends SunnyFormController<T> implements ModelBasedFormController<T> {
  late T _entity;

  @override
  T get mmodel => _entity;

  set mmodel(T mmodel) {
    assert(mmodel.mtype == mtype);
    this._entity = mmodel;
  }

  final MSchemaRef mtype;

  MBaseModelFormController(BuildContext context, this._entity,
      {required MSchemaRef? mtype,
      Map<JsonPath, FormFieldValidator>? validators})
      : mtype = mtype ?? _entity.mtype,
        super.base(context, validators ?? {}, "${mtype.baseCode}");

  @override
  void _set(JsonPath path, value) {
    return _entity.setByPath(path, value);
  }

  @override
  dynamic _get(JsonPath? path) {
    return _entity.getByPath(path!);
  }

  @override
  void _put(JsonPath path, value) {
    _entity.setByPath(path, value);
  }
}

extension MBaseModelDeep on MBaseModel {
  T? getByPath<T>(JsonPath path) {
    dynamic value = this;
    for (var segment in path.segments) {
      if (value is MModel) {
        value = value[segment];
      } else if (value is MBaseModel) {
        value = value[segment];
      } else if (value is Map) {
        value = value[segment];
      } else {
        throw Exception(
            "Illegal path: $path at segment $segment.  Expected Map or MModel but found ${value.runtimeType}");
      }
      if (value == null) {
        return null;
      }
    }
    return value as T?;
  }

  dynamic setByPath<T>(JsonPath<T> path, T? value) {
    if (value == "") value = null;
    final lastSegment = path.last;
    final parents = path.chop;
    dynamic container = this;
    for (var segment in parents.segments) {
      container = container[segment];
      if (container == null) {
        throw Exception(
            "Missing container in heirarchy.  Full path: $path.  Error found at segment $segment");
      }
    }
    if (value == null && container is Map) {
      container.remove(lastSegment);
    } else {
      container[lastSegment] = value;
    }
  }
}

class MModelForm<T extends MBaseModel> extends SunnyFormController<T>
    implements ModelBasedFormController<T> {
  late T _mmodel;

  @override
  T get mmodel => _mmodel;

  set mmodel(T mmodel) {
    assert(mmodel.mtype == mtype);
    this._mmodel = mmodel;
  }

  final MSchemaRef mtype;

  MModelForm(BuildContext context, this._mmodel,
      {required MSchemaRef? mtype,
      Map<JsonPath, FormFieldValidator>? validators})
      : mtype = mtype ?? _mmodel.mtype,
        super.base(context, validators ?? {}, "${mtype.baseCode}");

  @override
  void _set(JsonPath path, value) {
    return _mmodel.setByPath(path, value);
  }

  @override
  dynamic _get(JsonPath? path) {
    return _mmodel.getByPath(path!);
  }

  @override
  void _put(JsonPath path, value) {
    final map = <String, dynamic>{};
    Maps.setByPath(map, path, value);
    _mmodel.takeFromMap(map, copyEntries: true);
  }
}

class MSchemaForm<T extends MBaseModel> extends MModelForm<T> {
  final IMEntityDefinition? definition;

  MSchemaForm(
    BuildContext context,
    T model,
    IMSchemaDefinition? definition, {
    Map<JsonPath, FormFieldValidator>? validators,
    MSchemaRef? mtype,
  })  : definition =
            (definition ?? (mtype ?? model.mtype).getMSchemaOrNull(context))
                .asEntitySchema(),
        super(
          context,
          model,
          mtype: mtype ??
              (definition ?? (mtype ?? model.mtype).getMSchemaOrNull(context))
                  .asEntitySchema()
                  .schemaRef!,
          validators: validators,
        );
}

enum AttributeSource { model, control, unknown }

extension AttributeSourceExtention on AttributeSource {
  bool get isModel => this == AttributeSource.model;

  bool get isControl => this == AttributeSource.control;

  bool get isUnknown => this == AttributeSource.unknown;
}

class Attribute<X> {
  final JsonPath<X>? path;
  final X? value;
  final dynamic error;

  /// Used for the model to signal back to the control that an error has been cleared
  final bool clearErrors;
  final AttributeSource source;

  Attribute(this.path, this.value,
      {required AttributeSource source, this.clearErrors = false})
      : source = source,
        error = null;

  Attribute.control(JsonPath<X> path, X value)
      : this(path, value, source: AttributeSource.control);

  Attribute.model(JsonPath<X> path, X value)
      : this(path, value, source: AttributeSource.model);

  Attribute.ofError(this.path, this.error, {required AttributeSource source})
      : source = source,
        clearErrors = false,
        value = null;

  bool get isError => error != null;

  @override
  bool operator ==(Object other) =>
      identical(this, other) ||
      other is Attribute &&
          runtimeType == other.runtimeType &&
          path == other.path &&
          value == other.value;

  @override
  String toString() {
    return 'Attribute{path: $path, value: $value}';
  }

  @override
  int get hashCode => path.hashCode ^ value.hashCode;

  Attribute<C> cast<C>() {
    assert(value == null || value is C);
    return value == null
        ? Attribute<C>(path!.cast<C>(), null,
            source: source, clearErrors: clearErrors)
        : Attribute<C>(path!.cast<C>(), value as C?,
            source: source, clearErrors: clearErrors);
  }
}

extension FormControllerExtensions<T> on FormController<T?> {
  FutureOr<FormSubmitResult<X>> submitRoot<X>(BuildContext context,
      {bool? skipIfUnmodified, FutureOr<X>? submit()?}) async {
    FormController self = this;
    while (self.parent != null) {
      self = self.parent!;
    }

    return await self.submit(context,
        submit: submit, skipIfUnmodified: skipIfUnmodified);
  }

  FutureOr<FormSubmitResult<X>> submitIfModified<X>(BuildContext context,
      [FutureOr<X> submit()?]) async {
    return this.submit(context, submit: submit, skipIfUnmodified: true);
  }

//  MetaFormContext toMetaFormContext(
//      {Map<String, dynamic> scope,
//      MetadataOverrides overrides,
//      List<MetaPropertyHandler> extraHandlers,
//      FutureOr<MEntityDefinition> definition}) {
//    if (this is MModelForm) {
//      final form = this as MModelForm;
//      definition ??= mschemaService.getEntityDefinition(form.mtype);
//      return MetaFormContext.ofModelForm(
//        controller: form,
//        definition: definition,
//        scope: scope,
//        extraHandlers: extraHandlers,
//        overrides: overrides,
//      );
//    }
//    throw UnimplementedError("Not implemented");
//  }
}

typedef FormControllerCreator = FormController Function(BuildContext context);
typedef WidgetBuilderWithForm = Widget Function(
    BuildContext context, FormController controller);
typedef SubmitHo<T> = FutureOr<T> Function(FormController controller);

/// How you bind a [FormController] into the widget tree.  This class connects the form automatically to
/// any [Command] instances up the tree.
class SunnyForm<T> extends StatelessWidget {
  SunnyForm(
    this.name, {

    /// A builder for the form instance
    required this.form,
    required Widget this.child,
    this.isActive = true,
    this.isSliver = false,
    this.bindCommand,

    /// Code to execute when the form submits.  This essentially registers a callback hook to this form
    this.submitHook,

    /// A submit hook that runs during the validation phase
    this.validate,
    this.skipIfUnmodified,
  })  : builder = null,
        super(key: Key("sunny-form-$name"));

  SunnyForm.builder(
    this.name, {
    FormControllerCreator? form,
    required this.builder,
    this.isActive = true,
    this.skipIfUnmodified,
    this.bindCommand,
    this.isSliver = false,
    this.validate,
    this.submitHook,
  })  : assert(builder != null),
        child = null,
        form = form ??
            ((context) {
              return FormController.ofMap(context, {}, name: name);
            }),
        super(key: Key("sunny-form-$name"));

  final FormValidator? validate;
  final SubmitHook? submitHook;
  final FormControllerCreator form;
  final Widget? child;
  final WidgetBuilderWithForm? builder;
  final String name;
  final bool? skipIfUnmodified;
  final bool isActive;
  final bool isSliver;
  final bool? bindCommand;

  @override
  Widget build(BuildContext context) {
    return Provider<FormController>(
      key: Key("controller-$name-provider"),
      create: (context) {
        final _form = form(context);

        if (submitHook != null) {
          _form.registerSubmitHook(this.name, submitHook);
        }
        // Look for parent forms and register ourselves
        final parent = context
            .formOrNull(listen: false)
            ?.also((parent) => parent.registerForm(_form));

        if (parent == null && isActive && bindCommand == true) {
          // This call will also add unregister hooks
          _form.bindCommand(context, skipIfUnmodified: skipIfUnmodified);
        }

        return _form;
      },
      dispose: (_, form) {
        form.dispose();
      },
      child: Consumer<FormController>(
        builder: (context, form, child) {
          return FormAutofocus(
              form: form,
              child: (context) =>
                  child ?? builder?.call(context, form) ?? Container());
        },
        child: child,
      ),
    );
  }

  static FormController? of(BuildContext context,
      {bool listen = false, bool nullOk = false}) {
    FormController getter() => Provider.of(context, listen: listen);
    try {
      return getter();
    } catch (e) {
      if (!nullOk) {
        rethrow;
      } else {
        return null;
      }
    }
  }
}

class FormAutofocus extends StatefulWidget {
  final FormController form;
  final WidgetBuilder child;

  FormAutofocus({Key? key, required this.form, required this.child})
      : super(key: key ?? Key("form-autofocus-${form.name}"));

  @override
  _FormAutofocusState createState() => _FormAutofocusState();
}

class _FormAutofocusState extends State<FormAutofocus> with LoggingMixin {
  FocusScope? focusScope;

  @override
  Widget build(BuildContext context) {
    return widget.child(context);
  }

  @override
  void initState() {
    super.initState();
    Future.microtask(() {
      widget.form.autofocus();
    });
  }
}

extension MModelFormExtensions on MModel {
  MModelForm toSunnyForm(BuildContext context,
      {required MSchemaRef mtype,
      Map<JsonPath, FormFieldValidator>? validators}) {
    return MModelForm(context, this, mtype: mtype, validators: validators);
  }
}

extension AttributeStreamExt<X> on Stream<Attribute<X?>> {
  Stream<Attribute> operator [](JsonPath path) =>
      where((attr) => attr.path == path);

  Stream<Attribute<X>> get<X>(JsonPath<X>? path) =>
      where((attr) => attr.path == path).map((a) => a.cast<X>());

  Stream<X?> get control =>
      where((a) => a.source.isControl).map((a) => a.value);

  Stream<X?> get model => where((a) => a.source.isModel).map((a) => a.value);
}

/// In flutter, it can be challenging to bind submit buttons to forms, when the submit button lives outside the form's
/// tree.  These extensions make it easy to bind a form's submission to a `Command` that's registered further up the tree.
extension FormControllerCommandExtensions on FormController {
  /// Binds the command - raises an exception if no Command is found
  void bindCommand<T>(BuildContext context,
      {FutureOr<T> submit()?, bool? skipIfUnmodified = false}) {
    context.onBridge(this, () {
      return this.submitRoot(
        context,
        submit: submit ?? (() => null),
        skipIfUnmodified: skipIfUnmodified,
      );
    });
  }

  void unbindCommand<T>(CallBridge<T> command) {
    command.unregisterCall(this);
  }

  bool tryBindCommand<T>(BuildContext context,
      {FutureOr<T> submit()?, bool? skipIfUnmodified}) {
    try {
      var bridge = context.bridge<FormSubmitResult>();
      bridge?.also((command) {
        command.onEvent(this, () {
          return this.submitRoot<T>(
            context,
            submit: submit ?? (() => null),
            skipIfUnmodified: skipIfUnmodified,
          );
        });
        this.registerDisposer(() {
          this.unbindCommand(command);
        });
      });
      return true;
    } on ProviderNotFoundException {
      log.info("No binding for $name");
      return false;
    }
  }
}

extension BuildContextFormExtensions on BuildContext {
  FormController form({bool listen = false}) {
    return Provider.of(this, listen: listen);
  }

//  MetaFormContext get formMetadata {
//    return Provider.of(this);
//  }
//
//  MetaFormContext get formMetadataOrNull {
//    try {
//      return Provider.of(this);
//    } on ProviderNotFoundException {
//      return null;
//    }
//  }

  FormController? formOrNull({bool listen = true}) {
    try {
      return Provider.of(this, listen: listen);
    } on ProviderNotFoundException {
      return null;
    }
  }
}

extension ResultFuture<T> on FutureOr<Result<T>> {
  FutureOr<T> get value {
    return thenOr((_) => _.value);
  }
}

extension FormResultFuture<T> on FutureOr<FormSubmitResult<T>> {
  FutureOr<T> get value {
    return thenOr(((_) => _.value!) as T Function(FormSubmitResult<T>));
  }
}

class FormSubmitResult<X> with LoggingMixin {
  final FormController form;
  final Map<String, FormSubmitResult> subforms;
  final Map<String, Result> submitHooks;
  final Map<JsonPath, List<ValidationError>> _myValidationErrors;
  final X? value;
  final String? message;
  final ResultType? _resultType;
  final DateTime start;
  final DateTime end;
  ResultType? _calculatedResultType;
  List<dynamic>? _calculatedResults;
  Map<JsonPath, List<ValidationError>>? _calculatedErrors;

  FormSubmitResult(
      this.form,
      this.subforms,
      this._myValidationErrors,
      this.submitHooks,
      this.value,
      this.message,
      this._resultType,
      this.start,
      this.end);

  Map<JsonPath, List<ValidationError>>? get validationErrors {
    if (_calculatedErrors == null) {
      final flattenedErrors = <ValidationError>[
        ..._myValidationErrors.values.flatten(),
        ...subforms.values.expand((result) {
          return result.validationErrors!.values.flatten();
        }),
      ];
      _calculatedErrors = flattenedErrors.groupBy((err) => err.path);
    }
    return _calculatedErrors;
  }

  List<dynamic> get allResults {
    return _calculatedResults ??= [
      for (final form in this.subforms.values) ...form.allResults,
      for (final hook in this.submitHooks.values)
        if (hook.isSuccessful && hook.value != null) hook.value,
      if (value != null) value,
    ];
  }

  MModel? operator [](MSchemaRef ref) {
    final results = allResults.whereType<MModel>().where((_) => _.mtype == ref);
    if (results.isEmpty) {
      log.warning(
          "No result found with type ${ref.baseCode} ${allResults.isEmpty ? '(the results were empty)' : '- but we found ${allResults.map((_) => _.resultType)}'}");
      return null;
    } else if (results.length > 1) {
      return illegalState(
          "Found more than one result for ${ref.baseCode}, but found ${results.length}");
    } else {
      return results.first;
    }
  }

  T result<T>([bool filter(T element)?]) {
    final results =
        allResults.whereType<T>().where((_) => filter?.call(_) ?? true);
    if (results.isEmpty) {
      return illegalState(
          "No result found with type $T ${allResults.isEmpty ? '(the results were empty)' : '- but we found ${allResults.map((_) => _.resultType)}'}");
    } else if (results.length > 1) {
      return illegalState(
          "Found more than one result for $T, but found ${results.length}");
    } else {
      return results.first;
    }
  }

  T? resultOrNull<T>() {
    return allResults.whereType<T>().firstOrNull;
  }

  List<T> results<T>() {
    return allResults.whereType<T>().toList();
  }

  ResultType? get resultType {
    if (_calculatedResultType == null) {
      final results = [
        _resultType,
        ...subforms.values.map((_) => _.resultType)
      ];
      _calculatedResultType = results.firstWhere(
          (result) => result != ResultType.success,
          orElse: () => ResultType.success);
    }
    return _calculatedResultType;
  }

  @override
  String toString() {
    return buildString((str) {
      str += "FormSubmitResult{ \n";
      str += "  result: $resultType\n";
      final duration = (end.difference(start));
      str += "  duration: ${duration.format()}\n";
      if (message != null) str += message;
      if (subforms.isNotEmpty) str += "  subforms => \n";
      subforms.forEach((key, result) {
        str += "    $key: ${result.resultType}\n";
      });
      if (submitHooks.isNotEmpty) str += "  submitHooks => \n";
      submitHooks.forEach((key, result) {
        str += "    $key: ${result.resultType}\n";
      });
      if (allResults.isNotEmpty) str += "  model => \n";
      allResults.forEach((value) {
        str += "    ${value.runtimeType}: $value\n";
      });

      str += "}";
    });
  } //  @override
//  Result<X> withMessage(String message) {
//    return FormSubmitResult<X>(subforms, submitHooks, value, message, _resultType, start, end);
//  }

  FormSubmitResult.error(this.form, error)
      : subforms = {},
        submitHooks = {},
        start = DateTime.now(),
        _myValidationErrors = {},
        end = DateTime.now(),
        _resultType = ResultType.error,
        value = null,
        message = "$error";
}

extension FormSubmitResultExtensions<X> on FormSubmitResult<X> {
  bool get isSuccessful => this.resultType == ResultType.success;

  bool get isNotError => !isError;

  bool get isError => this.resultType == ResultType.error;

  bool get isNotSuccessful => this.resultType != ResultType.success;
}

class FormSubmitResultBuilder<X> {
  final FormController form;
  Map<String, FormSubmitResult> subforms = {};
  Map<String, Result> submitHooks = {};
  X? value;
  String? message;
  ResultType? resultType;
  final DateTime start;
  Map<JsonPath, List<ValidationError>> validationErrors = {};

  FormSubmitResultBuilder.start(this.form) : start = DateTime.now();

  FormSubmitResult<X> end() {
    return FormSubmitResult<X>(
      form,
      subforms,
      validationErrors,
      submitHooks,
      value,
      message,
      resultType,
      start,
      DateTime.now(),
    );
  }
}
