import 'dart:async';

import 'package:flutter/material.dart';
import 'package:sunny_platform_widgets/sunny_platform_widgets.dart';
import 'package:overlay_support/overlay_support.dart';
import 'package:sunny_dart/sunny_dart.dart';
import 'package:sunny_forms/error_message.dart';
import 'package:sunny_forms/show_notification.dart';
import 'package:sunny_forms/text_widgets.dart';

import 'color_extensions.dart';

typedef ToastPlacer = Widget Function(Widget inner);

void _closeEntryCallback(BuildContext context, OverlaySupportEntry? entry) {
  entry!.dismiss();
}

/// Class that assists in displaying overlays
class SunnyHud {
  static PlatformHud platformHud = !infoX.isAndroid ? IosHud2() : AndroidHud();

  static toast(BuildContext context, String text, {Duration? duration}) =>
      platformHud.toast(context, text, duration: duration);

  static notification(BuildContext context, String text,
          {bool isDismissible = true,
          OverlaySupportCallback onTap = _closeEntryCallback,
          Duration? duration}) =>
      showNotification(
        title: text,
        isSwipeDismissable: true,
        showDismissLink: isDismissible,
        duration: duration ?? 5.seconds,
        onTap: (_, entry) {
          onTap.call(_, entry);
        },
      );

  static Hud<T?> spinner<T>(BuildContext context, String text,
      {Future<T> operation()?, Duration? duration}) {
    return platformHud.spinner<T?>(context, text, operation: () async {
      if (operation != null) {
        try {
          final result = await Future.wait(
              [operation(), Future.delayed(kDefaultSunnyHudDuration)]);
          return result.first as T?;
        } catch (e) {
          /// Awaiting here caused issues because it was dismissing before returning the value
          dismiss(context);
          return null;
        }
      } else {
        return null;
      }
    });
  }

  static Hud info(BuildContext context, String text,
          {double? height, Widget? leading, Duration? duration}) =>
      platformHud.info(
        context,
        text,
        duration: duration,
        leading: leading,
      );

  static Hud overlay(BuildContext context, Widget widget,
          {Duration? duration, ToastPlacement? alignment}) =>
      platformHud.overlay(
        context,
        widget,
        alignment: alignment,
        duration: duration,
      );

  static Hud success(BuildContext context, String text, {Duration? duration}) =>
      platformHud.success(context, text, duration: duration);

  static Hud error(BuildContext context, String text,
          {Duration duration = const Duration(seconds: 7)}) =>
      platformHud.error(context, text, duration: duration);

  static FutureOr dismiss(BuildContext context, {bool animate = false}) {
    for (final toast in huds) {
      toast.dismiss(animate: animate);
    }
  }
}

typedef OverlayDismisser = FutureOr Function(bool animate);

class Hud<T> {
  factory Hud.from(OverlaySupportEntry entry,
      {required Future<T> operation()?, required Duration? duration}) {
    if (operation != null) {
      return Hud.ofOperation(entry, operation: operation);
    } else {
      return Hud.ofDelay(entry, duration);
    }
  }

  Hud.ofFuture(OverlaySupportEntry entry, {required Future<T> future})
      : _dismisser = ((animate) {
          entry.dismiss(animate: animate);
        }) {
    _completer.future.then((_) {
      huds.remove(this);
    });

    entry.dismissed.then((_) {
      _completer.complete();
    });

    future.then((_) {
      _completer.complete(_);
      entry.dismiss(animate: true);
    });
  }

  Hud.ofDelay(OverlaySupportEntry entry, Duration? duration)
      : this.ofFuture(
          entry,
          future: Future.delayed(duration ?? kDefaultSunnyHudDuration),
        );

  Hud.ofOperation(OverlaySupportEntry entry, {required Future<T> operation()})
      : this.ofFuture(entry, future: operation());

  final SafeCompleter<T> _completer = SafeCompleter<T>();

  Future<T> get value => _completer.future;
  final OverlayDismisser _dismisser;

  FutureOr dismiss({bool animate = true}) {
    return _dismisser(animate);
  }
}

abstract class PlatformHud {
  Hud overlay(BuildContext context, Widget widget,
      {bool autoDismiss = true,
      ToastPlacement? alignment,
      double? height,
      Duration? duration});

  Hud toast(BuildContext context, String text, {Duration? duration});

  Hud<T> spinner<T>(BuildContext context, String text,
      {Future<T> operation()?});

  Hud info(BuildContext context, String text,
      {double? height, Widget? leading, Duration? duration});

  Hud success(BuildContext context, String text, {Duration? duration});

  Hud error(BuildContext context, String text, {Duration? duration});

//  FutureOr dismiss(
//    BuildContext context,
//  );
}

//class IosHud implements PlatformHud {
//  @override
//  toast(BuildContext context, String text, {int duration = 1500}) async {
//    spinner(context, text);
//    dismissAfter(context, Duration(milliseconds: duration));
//  }
//
//  @override
//  spinner(BuildContext context, String text, {Duration duration}) async {
//    await dismiss(context);
//    await Progresshud.showWithStatus(text);
//    dismissAfter(context, duration);
//  }
//
//  @override
//  info(BuildContext context, String text, {Duration duration}) async {
//    await dismiss(context);
//    await Progresshud.showInfoWithStatus(text);
//    dismissAfter(context, duration);
//  }
//
//  dismissAfter(BuildContext context, Duration duration) {
//    if (duration != null) {
//      Future.delayed(duration, () => dismiss(context));
//    }
//  }
//
//  @override
//  success(BuildContext context, String text, {Duration duration}) async {
//    await dismiss(context);
//    await Progresshud.showSuccessWithStatus(text);
//    dismissAfter(context, duration);
//  }
//
//  @override
//  error(BuildContext context, String text, {Duration duration}) async {
//    await dismiss(context);
//    await Progresshud.showErrorWithStatus(text);
//    dismissAfter(context, duration);
//  }
//
//  @override
//  dismiss(BuildContext context) async {
//    int count = 0;
//    while (count < 10 && await Progresshud.isVisible()) {
//      count = count + 1;
//      await Progresshud.dismiss();
//    }
//  }
//}

final kDefaultSunnyHudDuration = 2.seconds;

final huds = <Hud>[];

/// Android Hud
class AndroidHud implements PlatformHud {
  AndroidHud();

  @override
  Hud toast(BuildContext context, String text,
      {Duration? duration = const Duration(milliseconds: 1500)}) {
    return spinner(context, text);
  }

  @override
  Hud<T> spinner<T>(BuildContext context, String text,
      {Future<T> operation()?}) {
    final hud = _showToast<T>(
        context,
        PlatformListTile(
          leading: PlatformCircularProgressIndicator(),
          title: Text(text),
        ),
        operation: operation,
        autoDismiss: false);
    return hud;
  }

  Hud<T> _showToast<T>(BuildContext context, Widget child,
      {Future<T> operation()?,
      bool autoDismiss = true,
      double? height,
      Duration? duration}) {
    final entry = showSimpleNotification(
      height != null ? SizedBox(height: height, child: child) : child,
      autoDismiss: false,
      slideDismissDirection: DismissDirection.endToStart,
    );

    final hudDisplay =
        Hud.from(entry, duration: duration, operation: operation);
    huds.add(hudDisplay);

    if (autoDismiss) {
      Future.delayed(duration ?? kDefaultSunnyHudDuration,
          () => hudDisplay.dismiss(animate: true));
    }
    return hudDisplay;
  }

  @override
  Hud overlay(BuildContext context, Widget widget,
      {double? height,
      ToastPlacement? alignment,
      bool autoDismiss = true,
      Duration? duration}) {
    return _showToast(context, widget,
        height: height, autoDismiss: true, duration: duration ?? 5.seconds);
  }

  @override
  Hud info(BuildContext context, String text,
      {double? height, Widget? leading, Duration? duration}) {
    return _showToast(context, Text(text), duration: duration);
  }

  @override
  Hud success(BuildContext context, String text, {Duration? duration}) {
    return _showToast(
      context,
      PlatformListTile(
        leading: Icon(
          Icons.check_circle_outline,
          color: Colors.green,
        ),
        title: Text(text),
      ),
      duration: duration,
    );
  }

  @override
  Hud error(BuildContext context, String text, {Duration? duration}) {
    return _showToast(context, ErrorMessage(text));
  }
}

/// Android Hud
class IosHud2 implements PlatformHud {
  IosHud2();

  @override
  Hud toast(BuildContext context, String text, {Duration? duration}) {
    return overlay(context, ListTile(title: Text(text)), duration: duration);
  }

  @override
  Hud overlay(BuildContext context, Widget widget,
      {bool autoDismiss = true,
      ToastPlacement? alignment,
      double? height,
      Duration? duration}) {
    return _showToast(context, widget,
        height: height,
        alignment: alignment,
        autoDismiss: autoDismiss,
        duration: duration ?? 5.seconds);
  }

  @override
  Hud<T> spinner<T>(BuildContext context, String text,
      {Duration? duration, Future<T> operation()?}) {
    return _showToast<T>(
        context,
        Chip(
          elevation: 2,
          avatar: Padding(
            padding: EdgeInsets.all(3),
            child: CircularProgressIndicator(
              strokeWidth: 2,
              valueColor: AlwaysStoppedAnimation<Color>(Colors.white),
            ),
          ),
          label: Text(text),
          backgroundColor: Theme.of(context).primaryColor,
          labelStyle: TextStyle(color: Colors.white),
        ),
        operation: operation,
        height: 60,
        duration: duration ?? Duration(seconds: 5));
  }

  Hud<T> _showToast<T>(BuildContext context, Widget child,
      {double? height = 70,
      ToastPlacement? alignment,
      bool autoDismiss = true,
      Duration? duration,
      Future<T> operation()?}) {
    final entry = showOverlay(
      (context, anim) {
        alignment ??= ToastPlacement.top;
        return alignment!(child, height ?? 70, anim);
      },
      duration: duration ??
          (autoDismiss ? Duration(seconds: 10) : Duration(minutes: 1)),
    );

    final hud = Hud.from(entry, operation: operation, duration: duration);
    huds.add(hud);
    return hud;
  }

  @override
  Hud info(BuildContext context, String text,
      {double? height, Widget? leading, Duration? duration}) {
    return _showToast(
        context,
        PlatformListTile(
          minHeight: height ?? 70,
          color: CupertinoVisualStyle.instance.cardColor.resolved(context),
          leading: leading,
          title: Text(text),
        ),
        height: height ?? 70,
        duration: duration);
  }

  @override
  Hud success(BuildContext context, String text, {Duration? duration}) {
    return _showToast(
      context,
      PlatformListTile(
        leading: Icon(
          Icons.check_circle_outline,
          color: Colors.green,
        ),
        title: Text(text),
      ),
      duration: duration,
    );
  }

  @override
  Hud error(BuildContext context, String text, {Duration? duration}) {
    return _showToast(
      context,
      PlatformListTile(
        leading: Icon(
          Icons.warning,
          color: Colors.red,
        ),
        title: ErrorText(text),
      ),
    );
  }
}

enum ToastPlacement { center, top, bottom }

class StaticValue {
  final int value;

  const StaticValue(this.value);

  Duration get ms => Duration(milliseconds: value);

  Duration get seconds => Duration(seconds: value);

  Duration get minutes => Duration(minutes: value);
}

const twoHundred = StaticValue(200);
const oneHundred = StaticValue(200);
const fifty = StaticValue(50);
const oneHundredFifty = StaticValue(200);
const threeHundred = StaticValue(200);

const instant = Duration(seconds: 0);
const twoSeconds = Duration(seconds: 2);
const tenSeconds = Duration(seconds: 10);
const oneAndAHalfSeconds = Duration(milliseconds: 1500);
const oneSecond = Duration(seconds: 1);
const halfSecond = Duration(milliseconds: 500);
const threeHundredMilliseconds = Duration(milliseconds: 300);
const oneHundredFiftyMs = Duration(milliseconds: 150);
const twoHundredMilliseconds = Duration(milliseconds: 200);

const ten = StaticValue(10);
const nine = StaticValue(9);
const eight = StaticValue(8);
const seven = StaticValue(7);
const six = StaticValue(6);
const five = StaticValue(5);
const four = StaticValue(4);
const three = StaticValue(3);
const two = StaticValue(2);
const one = StaticValue(1);

extension ToastPlacementExt on ToastPlacement {
  Widget call(Widget child, double height, double anim) {
    final self = this;
    switch (self) {
      case ToastPlacement.top:
        return Column(
          children: [
            Container(
              margin: EdgeInsets.only(top: 50),
              height: height,
              child: Opacity(
                child: child,
                opacity: anim,
              ),
            ),
            Expanded(
              child: Container(),
            ),
          ],
        );
      case ToastPlacement.center:
        return Column(
          mainAxisSize: MainAxisSize.max,
          mainAxisAlignment: MainAxisAlignment.center,
          children: [
            Container(
              margin: EdgeInsets.only(top: 50),
              height: height,
              child: Opacity(
                child: child,
                opacity: anim,
              ),
            ),
          ],
        );

      case ToastPlacement.bottom:
        return Column(
          children: [
            Expanded(
              child: Container(),
            ),
            Container(
              margin: EdgeInsets.only(bottom: 50),
              height: height,
              child: Opacity(
                child: child,
                opacity: anim,
              ),
            ),
          ],
        );
    }
  }
}
