import 'package:flexidate/flexidate.dart';
import 'package:flutter/cupertino.dart';
import 'package:flutter/material.dart';
import 'package:flutter_typeahead/flutter_typeahead.dart';
import 'package:provider/provider.dart';
import 'package:sunny_core_widgets/core_widgets.dart';
import 'package:sunny_dart/helpers/logging_mixin.dart';
import 'package:sunny_dart/typedefs.dart';
import 'package:sunny_essentials/sunny_essentials.dart';
import 'package:sunny_forms/clear_input.dart';
import 'package:sunny_forms/date_typeahead.dart';
import 'package:sunny_forms/form_fields/typeaheads.dart';
import 'package:sunny_forms/raw_text_field.dart';
import 'package:sunny_forms/sunny_form_controller.dart';
import 'package:sunny_forms/sunny_form_field.dart';
import 'package:sunny_forms/widget_extensions.dart';

/// A duration typeahead - control-only - see TimeSpanFormField for all the chrome
class TimeSpanTypeaheadControl extends StatefulWidget {
  final FormValueChanged<TimeSpan?>? onChange;
  final TimeSpan? initialValue;
  final String? placeholder;
  final String? label;
  final bool autofocus;
  final TextInputAction? textInputAction;
  final TextEditingController? textController;
  final bool enabled;
  final FocusNode? focusNode;
  final Formatter<TimeSpan?>? formatter;
  final bool? showIcon;

  const TimeSpanTypeaheadControl({
    Key? key,
    this.onChange,
    this.showIcon,
    this.initialValue,
    this.placeholder,
    this.autofocus = false,
    this.label,
    this.enabled = true,
    this.textInputAction,
    this.focusNode,
    this.textController,
    this.formatter,
  }) : super(key: key);

  @override
  _TimeSpanTypeaheadControlState createState() =>
      _TimeSpanTypeaheadControlState();
}

class _TimeSpanTypeaheadControlState extends State<TimeSpanTypeaheadControl>
    with LoggingMixin {
  late TextEditingController _textController;

  @override
  void initState() {
    super.initState();

    final startValue = widget.initialValue?.format() ?? "";
    _textController = widget.textController ??
        TextEditingController.fromValue(
          TextEditingValue(
            text: startValue,
            selection: TextSelection.collapsed(offset: startValue.length),
          ),
        );
    widget.focusNode?.addListener(_focusChange);
  }

  void _focusChange() {
    if (widget.focusNode!.hasFocus) {
      _textController.selection = TextSelection(
          baseOffset: 0, extentOffset: _textController.text.length);
    }
  }

  @override
  void dispose() {
    if (widget.textController == null) {
      _textController.dispose();
    }
    widget.focusNode?.removeListener(_focusChange);
    super.dispose();
  }

  @override
  Widget build(BuildContext context) {
    final formatter =
        widget.formatter ?? ((timeSpan) => timeSpan?.format() ?? '');
    final textFieldConfig = getTypeaheadInputConfig(
      context,
      placeholder: widget.placeholder,
      inputAction: widget.textInputAction,
      focusNode: widget.focusNode,
      controller: _textController,
      autofocus: widget.autofocus,
      hasPrefix: widget.showIcon != false,
      hasSuffix: true,
    );

    var input = inputWithIcon(
      prefix: widget.showIcon != false ? context.icons.alarm.icon : null,
      suffix: ClearInputIcon(
        notifier: widget.onChange,
        focusNode: widget.focusNode,
        textEditingController: _textController,
        showExpand: false,
      ),
      color: widget.focusNode?.hasFocus == true || widget.initialValue != null
          ? sunnyColors.primaryColor
          : sunnyColors.textLight,
      input: TypeAheadField<TimeSpan>(
        autoFlipDirection: true,
        textFieldConfiguration: textFieldConfig,
        noItemsFoundBuilder: (context) {
          return buildNoItemsTile(context, message: "Enter a duration above");
        },
        suggestionsCallback: (query) async {
          final amount = double.tryParse(query.trim());
          late List<TimeSpan> spans;
          if (amount != null) {
            spans = [
              TimeSpan(seconds: amount.toInt()),
              TimeSpan(minutes: amount.toInt()),
              TimeSpan(hours: amount.toInt()),
            ];
          } else {
            spans = parseTimeSpans(query);
          }

          // if (spans.length == 1) {
          // widget.onChange?.call(spans.first, AttributeSource.model);
          // }
          return spans;
        },
        errorBuilder: typeaheadErrorBuilder,
        animationStart: 1.0,
        transitionBuilder: (context, widget, controller) {
          return widget;
        },
        itemBuilder: (BuildContext context, TimeSpan itemData) {
          return InputWidgets.buildTypeaheadOption(
              title: Text(formatter(itemData)));
        },
        hideOnLoading: false,
        hideOnEmpty: false,
        hideOnError: false,
        hideSuggestionsOnKeyboardHide: false,
        onSuggestionSelected: (TimeSpan suggestion) {
          widget.onChange!(suggestion, AttributeSource.control);
          _textController.text = formatter(suggestion);
          return;
        },
      ),
    );

    return widget.focusNode == null
        ? input
        : FocusTrapArea(focusNode: widget.focusNode!, child: input);
  }
}
