import 'dart:async';

import 'package:flutter/cupertino.dart';
import 'package:provider/provider.dart';
import 'package:sunny_dart/sunny_dart.dart';
import 'package:sunny_forms/builders/fisher.dart';
import 'package:sunny_forms/options/options_service.dart';
import 'package:sunny_forms/standard_column.dart';
import 'package:sunny_forms/sunny_form_controller.dart';
import 'package:sunny_forms/text_widgets.dart';
import 'package:sunny_service_stubs/models.dart';
import 'package:sunny_service_stubs/options.dart';
import 'package:sunny_service_stubs/services.dart';

typedef WidgetValueBuilder<T> = Widget Function(BuildContext context, T value);
typedef WidgetDualValueBuilder<T, T2> = Widget Function(
    BuildContext context, T value, T2 t2);

class WidgetExtensions {
  WidgetExtensions._();
}

extension StreamWidgetBuilder<T> on Stream<T> {
  Widget build(BuildContext context, HStreamBuilder<T> builder,
      {WidgetBuilder? loadingBuilder, T? initial, Key? key}) {
    return StreamBuilder<T>(
      key: key,
      builder: (context, snap) {
        final data = snap.data ?? initial;
        return data == null ? loadingBuilder!(context) : builder(context, data);
      },
      initialData: initial,
    );
  }
}

extension StringKeyExtensions on String? {
  Key? toKey([String? suffix]) =>
      this == null ? null : Key("${this}${suffix ?? ''}");
}

extension IconExtensions on Icon {
  Icon resized(double size, {Color? color}) {
    return Icon(this.icon, size: size, color: color ?? this.color);
  }

  Icon colored(Color color, {double? size}) {
    return Icon(this.icon, size: size ?? this.size, color: color);
  }
}

// extension DynamicColorExt on Color {
//   Color resolved(BuildContext context) {
//     final self = this;
//     if (self == null) return null;
//     if (self is CupertinoDynamicColor) return self.resolveFrom(context);
//
//     return self;
//   }
// }

extension OptionsHandlerExtension<K, T> on KeyedOptionsHandler<K, T> {
  /// Given a key [key], this will attempt to resolve it either by:
  /// a) loading the value using [loadValue]
  /// b) creating an ad-hoc option
  FutureOr<KeyedOption<K, T>?> resolveOption(final K? key) {
    if (key == null) return null;
    final loadedValue = this.loadValue(key);
    return key is! String
        ? loadedValue.thenOr(((value) => toOption(value)!))
        : loadedValue.thenOr((lookup) {
            if (lookup != null) return toOption(lookup)!;
            return adhocOptionBuilder?.call(key);
          });
  }

  List<FutureOr<KeyedOption<K, T>?>?> resolveAllOptions(
      BuildContext context, final Iterable<K>? inputs) {
    if (inputs == null) return [];
    return [...inputs.map((k) => resolveOption(k)).notNull()];
  }
}

extension SunnyControllerWidgetExtensions on FormController {
  Widget pathBuilder<T>(JsonPath path, WidgetValueBuilder<T?> builder) {
    return StreamBuilder<T?>(
      key: Key("observe-$path"),
      stream: this.observePath(path),
      builder: (context, snap) => builder(context, snap.data),
      initialData: this.get(path),
    );
  }

  Widget pathsBuilder<T, T2>(JsonPath first, JsonPath second,
      WidgetDualValueBuilder<T?, T2?> builder) {
    return pathBuilder<T>(first, (context, firstValue) {
      return pathBuilder<T2>(second, (context, secondValue) {
        return builder(context, firstValue, secondValue);
      });
    });
  }
}

extension DimensionsInsetsExt on IDimensions {
  EdgeInsets toEdgeInsets() {
    return EdgeInsets.symmetric(
      vertical: this.height?.toDouble() ?? 0.0,
      horizontal: this.width?.toDouble() ?? 0.0,
    );
  }
}

extension ValueNotifierExtension<T> on ValueNotifier<T> {
  StreamController<T> publish([bool isBroadcast = false]) {
    StreamController<T>? _controller;
    final _onNotify = () {
      _controller?.add(this.value);
    };
    final _onCancel = () {
      removeListener(_onNotify);
    };
    _controller = isBroadcast
        ? StreamController.broadcast(onCancel: _onCancel)
        : StreamController(onCancel: _onCancel);
    addListener(_onNotify);
    return _controller;
  }
}

extension ValueStreamWidgets<T> on ValueStream<T> {
  Widget build(String key, Widget builder(BuildContext context, T? t),
      {bool waitForFirst = true}) {
    return StreamBuilder<T>(
        key: Key(key),
        initialData: this.resolve(),
        stream: this.after,
        builder: (context, snapshot) {
          if (snapshot.hasData || waitForFirst == false) {
            return builder(context, snapshot.data);
          } else {
            return emptyBox;
          }
        });
  }
}

extension WidgetOffstateExtensions on Widget? {
  /// Converts a nullable widget to a builder
  WidgetBuilder? toBuilder() {
    return this == null ? null : (_) => this!;
  }

  Widget offstage([bool offStageIf = true, bool forceWrap = false]) {
    if (offStageIf == true || forceWrap == true) {
      return Offstage(offstage: offStageIf == true, child: this);
    } else {
      return this!;
    }
  }

  Widget _pad(
      {double? horizontal,
      double? vertical,
      double? amount,
      EdgeInsets? padding}) {
    if (amount != null) {
      return Padding(padding: EdgeInsets.all(amount), child: this);
    } else if (padding != null) {
      return Padding(padding: padding, child: this);
    } else if (horizontal != null || vertical != null) {
      horizontal ??= 0;
      vertical ??= 0;
      return Padding(
        padding: EdgeInsets.only(
            left: horizontal,
            right: horizontal,
            top: vertical,
            bottom: vertical),
        child: this,
      );
    } else {
      return this!;
    }
  }

  Widget tappable(VoidCallback? onTap,
      {double? padAmount, EdgeInsets? padding}) {
    if (onTap == null) {
      return this!;
    }
    return GestureDetector(
      onTap: onTap,
      behavior: HitTestBehavior.opaque,
      child: this._pad(amount: padAmount, padding: padding),
    );
  }
}

extension NamedIconExtensions on NamedIcon {
  Widget toListSelectOption() {
    return this.sized(25).widget;
  }

  Widget toPrefixIcon() {
    return this.sized(25).widget;
  }
}

extension ListToColumn on List? {
  Widget? toColumn([Widget ifEmpty()?]) {
    final columnItems = (this ?? [])

        /// We accept non blank strings and non-null everything elses
        .where((final _) =>
            (_ is! String && _ != null) || (_ is String && _.isNotNullOrBlank));

    if (columnItems.isEmpty) {
      return null;
    } else {
      return column([
        for (final line in columnItems) textOrNull(line)!,
      ]);
    }
  }
}

extension NumberWidgetExtensions on num {
  double orZero() {
    return this.toDouble();
  }

  EdgeInsets toPadding() {
    return EdgeInsets.all(this.toDouble());
  }

  EdgeInsets toTopPadding() {
    return EdgeInsets.only(top: this.toDouble());
  }

  EdgeInsets toHorizontalPadding() {
    return EdgeInsets.symmetric(horizontal: this.toDouble());
  }

  EdgeInsets toVerticalPadding() {
    return EdgeInsets.symmetric(vertical: this.toDouble());
  }
}

extension BuildContextProvides on BuildContext {
  T? consume<T>([void exec(T value)?]) {
    try {
      final t = Provider.of<T>(this, listen: false);
      exec?.call(t);
      return t;
    } on ProviderNotFoundException {
      return null;
    }
  }

  NamedIcon getIcon(String name) {
    final NamedIcon? icon = consume<IconResolver>()?.getIconOrNull(name)!;
    assert(icon != null, "Icon $name wasn't found");
    return icon!;
  }

  NamedIcon? findIcon(String name) {
    final icon = consume<IconResolver>()?.getIconOrNull(name);
    return icon;
  }

  NamedIconContainer get icons {
    return sunny.get<NamedIconContainer>();
  }

//  SunnyListController<T> listController<T>() {
//    try {
//      return Provider.of(this, listen: false);
//    } on ProviderNotFoundException {
//      return null;
//    }
//  }

//  void onPageRefresh(void callback()) {
//    try {
//      final refresh = Provider.of<PageRefreshNotifier>(this, listen: false);
//      refresh.addListener(callback);
//    } on ProviderNotFoundException {
//      _log.warning(
//          "Couldn't find page refresh - should use this inside a page");
//    }
//  }
}
